# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Counter
from typing import Tuple

import torch.fx

import nncf.torch.graph.operator_metatypes as om
from nncf.common.graph import NNCFNode
from nncf.common.graph.layer_attributes import BaseLayerAttributes
from nncf.common.graph.layer_attributes import ConvolutionLayerAttributes
from nncf.common.graph.layer_attributes import Dtype
from nncf.common.graph.operator_metatypes import UnknownMetatype
from nncf.common.logging import nncf_logger
from nncf.experimental.torch.fx.node_utils import get_tensor_constant_from_node
from nncf.torch.dynamic_graph.layer_attributes_handlers import apply_args_defaults
from nncf.torch.graph.graph import PTNNCFGraph
from nncf.torch.graph.operator_metatypes import PT_OPERATOR_METATYPES


class GraphConverter:
    """
    Builds the NNCFGraph from an torch.fx.GraphModule instance.
    """

    def _get_layer_attributes(
        node: torch.fx.Node, metatype: om.OperatorMetatype, model: torch.fx.GraphModule
    ) -> BaseLayerAttributes:
        """
        Collects layer attributes for the given node.

        :param node: Given node.
        :param metatype: Given node metatype.
        :param model: Target GraphModule instance.
        :return: Given node layer attributes.
        """
        if metatype in [om.PTConv1dMetatype, om.PTConv2dMetatype, om.PTConv3dMetatype]:
            conv_default_args = [(arg.name, arg.default_value) for arg in node.target._schema.arguments]
            kwargs = apply_args_defaults(node.args, node.kwargs, conv_default_args)

            weight_node = kwargs["weight"]
            if weight_node.op != "get_attr":
                # Convs with constant subgraphs or two inputs are not supported yet.
                return None
            weight = get_tensor_constant_from_node(weight_node, model)
            return ConvolutionLayerAttributes(
                weight_requires_grad=False,
                in_channels=weight.shape[0],
                out_channels=weight.shape[1],
                kernel_size=list(weight.shape[2:]),
                stride=kwargs["stride"],
                dilations=kwargs["dilation"],
                groups=kwargs["groups"],
                padding_values=kwargs["padding"],
                transpose=False,
            )
        return None

    def _map_fx_unique_metatypes(node: torch.fx.Node, metatype: om.OperatorMetatype) -> om.OperatorMetatype:
        """
        Attempts to retrieve correct subtype for the given node.

        :param node: Given node.
        :param metatype: Given node metatype.
        :param model: Target GraphModule instance.
        :return: Correct FX metatype of the given node if it is exist or the original node metatype otherwise.
        """
        if metatype in [om.PTEmbeddingMetatype]:
            weight_node = node.args[0]
            if weight_node.op == "get_attr":
                return om.PTAtenEmbeddingMetatype

        return metatype

    @staticmethod
    def get_node_type_and_metatype(node: torch.fx.Node, model: torch.fx.GraphModule) -> Tuple[str, om.OperatorMetatype]:
        """
        Retrieves node's type and metatype.

        :param node: Given node.
        :param model: Given GraphModule.
        :return: Node's type and metatype.
        """
        if node.op == "placeholder":
            node_type = "input"
            node_metatype = om.PTInputNoopMetatype
        elif node.op == "output":
            node_type = "output"
            node_metatype = om.PTOutputNoopMetatype
        elif node.op == "get_attr":
            node_type = "get_attr"
            node_metatype = om.PTConstNoopMetatype
        elif node.op in ("call_function",):
            if hasattr(node.target, "overloadpacket"):
                node_type = str(node.target.overloadpacket).split(".")[1]
            elif node.target.__name__ == "getitem":
                node_type = "__getitem__"
            else:
                # TODO(dlyakhov): get correct nodes types from this nodes as well
                node_type = str(node.target)
            node_metatype = PT_OPERATOR_METATYPES.get_operator_metatype_by_op_name(node_type)
        else:
            node_type = node.op
            node_metatype = UnknownMetatype
        if node_metatype is UnknownMetatype:
            nncf_logger.debug(f"Unknown metatype for node: {node}")

        if node_metatype.get_subtypes():
            layer_attrs = GraphConverter._get_layer_attributes(node, node_metatype, model)
            node_subtype = node_metatype.determine_subtype(layer_attrs)
            node_metatype = node_subtype or node_metatype
        return node_type, node_metatype

    @staticmethod
    def create_nncf_graph(model: torch.fx.GraphModule) -> PTNNCFGraph:
        """
        Creates NNCFGraph from GraphModule.
        All nodes from model which have valid metatype are added to NNCFGraph.
        Then, corresponding edges are added to the NNCFGraph with shape, type, output and input port ids.

        :param model: torch fx GraphModule.
        :return: NNCFGraph.
        """
        nncf_graph = PTNNCFGraph()

        const_targets_counter = Counter([node.target for node in model.graph.nodes if node.op == "get_attr"])
        for source_node in model.graph.nodes:
            node_type, node_metatype = GraphConverter.get_node_type_and_metatype(source_node, model)
            node_metatype = GraphConverter._map_fx_unique_metatypes(source_node, node_metatype)
            is_shared_node = source_node.op in ("get_attr",) and (
                const_targets_counter[source_node.target] > 1 or len(source_node.users) > 1
            )

            nncf_graph.add_nncf_node(
                node_name=source_node.name, node_type=node_type, node_metatype=node_metatype, is_shared=is_shared_node
            )

        for source_node in model.graph.nodes:
            source_nncf_node = nncf_graph.get_node_by_name(source_node.name)
            for idx, dist_node in enumerate(source_node.users):
                dist_node_id = nncf_graph.get_node_by_name(dist_node.name).node_id
                input_port_id, output_port_id, tensor_shape = GraphConverter.get_edge_params(
                    model, source_node, source_nncf_node, dist_node, idx
                )
                nncf_graph.add_edge_between_nncf_nodes(
                    source_nncf_node.node_id,
                    dist_node_id,
                    tensor_shape=tensor_shape,
                    input_port_id=input_port_id,
                    output_port_id=output_port_id,
                    dtype=Dtype.FLOAT,
                )
        return nncf_graph

    @staticmethod
    def get_edge_params(
        model: torch.fx.GraphModule,
        source_node: torch.fx.Node,
        source_nncf_node: NNCFNode,
        dist_node: torch.fx.Node,
        output_idx: int,
    ) -> Tuple[int, int, Tuple[int, ...]]:
        """
        Retrieves edge params from the given source_node and dist_node pair.

        :param model: A torch.fx.GraphModule instance.
        :param source_node: Source node in format of torch.fx.Node.
        :param source_nncf_node: Source node in format of NNCFNode.
        :param dist_node: Distance node in format of torch.fx.Node.
        :param output_idx: Output index of the source_node.
        :return: Tuple of edge parameters: edge input port id, edge output port id and
            edge tensor shape.
        """
        output_port_id = 0
        tensor_shape = None
        if source_node.op in ("get_attr",):
            tensor_shape = tuple(get_tensor_constant_from_node(source_node, model).shape)
        elif "val" in source_node.meta:
            if source_nncf_node.metatype is om.PTBatchNormMetatype and isinstance(
                source_node.meta["val"], (tuple, list)
            ):
                tensor = source_node.meta["val"][0]
            elif source_nncf_node.metatype in [om.PTSplitMetatype, om.PTMaxMetatype, om.PTMinMetatype]:
                tensor = source_node.meta["val"][output_idx]
                # Assume every outputs corresponds to an unique output_port_id
                output_port_id = output_idx
            else:
                tensor = source_node.meta["val"]
            if isinstance(tensor, torch.Tensor):
                tensor_shape = tuple(tensor.shape)

        if tensor_shape is None:
            # TODO(dlyakhov): Refactor algorithms to always have knowns edges shapes.
            nncf_logger.debug(f"Edge shape between {source_node.name} and {dist_node.name} is unknown.")

        input_port_id = dist_node.all_input_nodes.index(source_node)
        return input_port_id, output_port_id, tensor_shape



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional, Tuple, cast

import numpy as np
import scipy.optimize  # type: ignore[import-untyped]

from nncf.api.compression import CompressionAlgorithmController
from nncf.common.schedulers import BaseCompressionScheduler
from nncf.common.schedulers import ExponentialDecaySchedule
from nncf.common.utils.api_marker import api
from nncf.common.utils.registry import Registry
from nncf.config.schemata.defaults import PRUNING_NUM_INIT_STEPS
from nncf.config.schemata.defaults import PRUNING_STEPS
from nncf.config.schemata.defaults import PRUNING_TARGET

PRUNING_SCHEDULERS = Registry("pruning_schedulers")


@api()
class PruningScheduler(BaseCompressionScheduler):
    """
    This is the class from which all pruning schedulers inherit.

    A pruning scheduler is an object which specifies the pruning
    level at each training epoch. It involves a scheduling algorithm,
    defined in the `_calculate_pruning_level()` method and a state
    (some parameters required for current pruning level calculation)
    defined in the `__init__()` method.

    :param controller: Pruning algorithm controller.
    :param params: Parameters of the scheduler in the JSON-like dictionary form. Passed as-is from the corresponding
      section of the NNCF config file .json section (https://openvinotoolkit.github.io/nncf/schema).
    """

    def __init__(self, controller: CompressionAlgorithmController, params: Dict[str, Any]):
        super().__init__()
        self._controller = controller
        self.initial_level = self._controller.pruning_init  # type: ignore[attr-defined]

        if self._controller.prune_flops:  # type: ignore[attr-defined]
            self.target_level = cast(float, params.get("pruning_flops_target"))
        else:
            self.target_level = cast(float, params.get("pruning_target", PRUNING_TARGET))

        self.num_warmup_epochs = cast(int, params.get("num_init_steps", PRUNING_NUM_INIT_STEPS))
        self.num_pruning_epochs = cast(int, params.get("pruning_steps", PRUNING_STEPS))
        self.freeze_epoch = self.num_warmup_epochs + self.num_pruning_epochs

    def _calculate_pruning_level(self) -> float:
        """
        Calculates a pruning level that should be applied to the model
        for the `current_epoch`.

        :return: Pruning level that should be applied to the model.
        """
        raise NotImplementedError("PruningScheduler implementation must override _calculate_pruning_level method.")

    def epoch_step(self, next_epoch: Optional[int] = None) -> None:
        """
        Should be called at the beginning of each training epoch to prepare
        the pruning method to continue training the model in the `next_epoch`.

        :param next_epoch: The epoch index for which the pruning scheduler
            will update the state of the pruning method.
        """
        super().epoch_step(next_epoch)
        self._controller.set_pruning_level(self.current_pruning_level)  # type: ignore[attr-defined]
        if self.current_epoch >= self.freeze_epoch:
            self._controller.freeze()  # type: ignore[attr-defined]

    def step(self, next_step: Optional[int] = None) -> None:
        """
        Should be called at the beginning of each training step to prepare
        the pruning method to continue training the model in the `next_step`.

        :param next_step: The global step index for which the pruning scheduler
            will update the state of the pruning method.
        """
        super().step(next_step)
        self._controller.step(next_step)  # type: ignore[attr-defined]

    @property
    def current_pruning_level(self) -> float:
        """
        Returns pruning level for the `current_epoch`.

        :return: Current sparsity level.
        """
        if self.current_epoch >= self.num_warmup_epochs:
            return self._calculate_pruning_level()
        return 0


@PRUNING_SCHEDULERS.register("baseline")
class BaselinePruningScheduler(PruningScheduler):
    """
    Pruning scheduler which applies the same pruning level for each epoch.

    The model is trained without pruning during `num_warmup_epochs` epochs.
    Then scheduler sets `target_level` and freezes the algorithm.
    """

    def __init__(self, controller: CompressionAlgorithmController, params: Dict[str, Any]):
        super().__init__(controller, params)
        self.freeze_epoch = self.num_warmup_epochs

    def _calculate_pruning_level(self) -> float:
        return self.target_level


@PRUNING_SCHEDULERS.register("exponential")
class ExponentialPruningScheduler(PruningScheduler):
    """
    Pruning scheduler with an exponential decay schedule.

    This scheduler applies exponential decay to the density level
    to calculate the pruning level for the `current_epoch`.
    The density level for the `current_epoch` is calculated as

        current_density = 1.0 - current_level
    """

    def __init__(self, controller: CompressionAlgorithmController, params: Dict[str, Any]):
        """
        Initializes a pruning scheduler with an exponential decay schedule.

        :param controller: Sparsity algorithm controller.
        :param params: Parameters of the scheduler.
        """
        super().__init__(controller, params)
        initial_density = 1.0 - self.initial_level
        target_density = 1.0 - self.target_level
        target_epoch = self.num_pruning_epochs - 1
        self.schedule = ExponentialDecaySchedule(initial_density, target_density, target_epoch)

    def _calculate_pruning_level(self) -> float:
        current_density = self.schedule(self.current_epoch - self.num_warmup_epochs)
        current_level = 1.0 - current_density
        return min(current_level, self.target_level)


@PRUNING_SCHEDULERS.register("exponential_with_bias")
class ExponentialWithBiasPruningScheduler(PruningScheduler):
    """
    Pruning scheduler which calculates pruning level for the current epoch
    according to the formula:

        current_level = a * exp(-k * epoch_idx) + b,

    where a, b, k is a params.
    """

    def __init__(self, controller: CompressionAlgorithmController, params: Dict[str, Any]):
        """
        Initializes a pruning scheduler with an exponential (with bias) decay schedule.

        :param controller: Sparsity algorithm controller.
        :param params: Parameters of the scheduler.
        """
        super().__init__(controller, params)
        target_epoch = self.num_pruning_epochs - 1
        self.a, self.b, self.k = self._init_exp(target_epoch, self.initial_level, self.target_level)

    def _calculate_pruning_level(self) -> float:
        current_level = self.a * float(np.exp(-self.k * (self.current_epoch - self.num_warmup_epochs))) + self.b
        return min(current_level, self.target_level)

    @staticmethod
    def _init_exp(epoch_idx: int, p_min: float, p_max: float, factor: float = 0.125) -> Tuple[float, float, float]:
        """
        Finds parameters a, b, k from the system:
            p_min = a + b
            p_max = a * exp(-k * epoch_idx) + b
            3/4 * p_max = a * exp(-k * factor * epoch_idx) + b
        For more details see [paper](https://arxiv.org/pdf/1808.07471.pdf).

        :param epoch_idx: Zero-based index of the epoch for which the a * exp(-k * epoch_idx) + b = p_max.
        :param p_min: Initial pruning level at which the schedule begins.
        :param p_max: Target pruning level at which the schedule ends.
        :param factor: Hyperparameter.
        """

        def get_b(a: float) -> float:
            return p_min - a

        def get_a(k: float) -> float:
            return (p_max - p_min) / (float(np.exp(-k * epoch_idx)) - 1)

        def f_to_solve(x: float) -> float:
            c = (0.75 * p_max - p_min) / (p_max - p_min)
            y = float(np.exp(-x * epoch_idx))
            return float(y**factor) - c * y + c - 1

        k = float(scipy.optimize.fsolve(f_to_solve, [1])[0])
        a = get_a(k)
        b = get_b(a)
        return a, b, k



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.fx

from nncf.common.graph.graph import NNCFGraph
from nncf.common.quantization.quantizer_setup import SingleConfigQuantizerSetup
from nncf.experimental.quantization.quantizer import Quantizer
from nncf.experimental.torch.fx.quantization.quantizer.openvino_quantizer import OpenVINOQuantizer


class OpenVINOQuantizerAdapter(Quantizer):
    """
    Implementation of the NNCF Quantizer interface for the OpenVINOQuantizer.
    """

    def __init__(self, quantizer: OpenVINOQuantizer):
        self._quantizer = quantizer

    def transform_prior_quantization(self, model: torch.fx.GraphModule) -> torch.fx.GraphModule:
        return self._quantizer.transform_for_annotation(model)

    def get_quantization_setup(self, model: torch.fx.GraphModule, nncf_graph: NNCFGraph) -> SingleConfigQuantizerSetup:
        return self._quantizer.get_quantization_setup(model, nncf_graph)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

import tensorflow as tf

from nncf.common.quantization.structs import QuantizationScheme as QuantizationMode
from nncf.common.utils.registry import Registry
from nncf.tensorflow.layers.operation import InputType
from nncf.tensorflow.quantization.quantizers import AsymmetricQuantizer
from nncf.tensorflow.quantization.quantizers import SymmetricQuantizer
from nncf.tensorflow.quantization.quantizers import TFQuantizerSpec

NNCF_QUANTIZATION_OPERATIONS_V2 = Registry("nncf_quantization_operations_v2")


@NNCF_QUANTIZATION_OPERATIONS_V2.register(QuantizationMode.SYMMETRIC)
class SymmetricQuantizerV2(SymmetricQuantizer):
    def set_input_spec(
        self, input_type: str, input_shape: Optional[List[int]] = None, channel_axes: Optional[List[int]] = None
    ):
        """
        Sets input tensor specification for the quantizer.

        :param input_type: Indicates the type of input tensor: `inputs` or `weights`.
        :param input_shape: Shape of the input tensor for which the
            quantization is applied. Required only for per-channel
            quantization.
        :param channel_axes: Axes numbers of the input tensor which
            correspond to its channels. Required only for per-channel
            quantization.
        """
        self.input_type = input_type
        self.input_shape = input_shape
        self.channel_axes = channel_axes

    def create_variables(self, layer: tf.keras.layers.Layer) -> Dict[str, tf.Variable]:
        """
        Creates quantizer variables using `layer.add_weight()` method.

        :param layer: Instance of the `tf.keras.layers.Layer` class.
        :return: Quantizer variables.
        """
        if self.per_channel and (self.input_shape is None or self.channel_axes is None):
            raise ValueError(
                "The `input_shape` and `channel_axes` arguments are required when using per-channel quantization."
            )
        prefix = self.name
        return self._create_variables(layer, self.input_shape, self.channel_axes, prefix)


@NNCF_QUANTIZATION_OPERATIONS_V2.register(QuantizationMode.ASYMMETRIC)
class AsymmetricQuantizerV2(AsymmetricQuantizer):
    def set_input_spec(
        self, input_type: str, input_shape: Optional[List[int]] = None, channel_axes: Optional[List[int]] = None
    ):
        """
        Sets input tensor specification for the quantizer.

        :param input_type: Indicates the type of input tensor: `inputs` or `weights`.
        :param input_shape: Shape of the input tensor for which the
            quantization is applied. Required only for per-channel
            quantization.
        :param channel_axes: Axes numbers of the input tensor which
            correspond to its channels. Required only for per-channel
            quantization.
        """
        self.input_type = input_type
        self.input_shape = input_shape
        self.channel_axes = channel_axes

    def create_variables(self, layer: tf.keras.layers.Layer) -> Dict[str, tf.Variable]:
        """
        Creates quantizer variables using `layer.add_weight()` method.

        :param layer: Instance of the `tf.keras.layers.Layer` class.
        :return: Quantizer variables.
        """
        if self.per_channel and (self.input_shape is None or self.channel_axes is None):
            raise ValueError(
                "The `input_shape` and `channel_axes` arguments are required when using per-channel quantization."
            )
        prefix = self.name
        return self._create_variables(layer, self.input_shape, self.channel_axes, prefix)


def create_quantizer(
    name: str,
    qspec: TFQuantizerSpec,
    is_weight_quantization: bool,
    input_shape: Optional[List[int]] = None,
    channel_axes: Optional[List[int]] = None,
):
    """
    Factory method to create quantizer.

    :param name: Name of the quantizer. Should be unique.
    :param qspec: Specification of the quantizer.
    :param is_weight_quantization: A boolean flag.
        Takes one of the following values:
            - `True` if input tensor of the quantizer is weights
            - `False` if input tensor of the quantizer is activations
    :param input_shape: Shape of the input tensor for which the
        quantization is applied. Required only for per-channel
        quantization.
    :param channel_axes: Axes numbers of the input tensor which
        correspond to its channels. Required only for per-channel
        quantization.
    :return: The instance of the `SymmetricQuantizerV2` or
        `AsymmetricQuantizerV2` class.
    """
    quantizer_cls = NNCF_QUANTIZATION_OPERATIONS_V2.get(qspec.mode)
    input_type = InputType.WEIGHTS if is_weight_quantization else InputType.INPUTS
    quantizer = quantizer_cls(name, qspec)
    quantizer.set_input_spec(input_type, input_shape, channel_axes)
    return quantizer



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Dict, Generic, Hashable, List, Optional, TypeVar

T = TypeVar("T")


class Cluster(Generic[T]):
    """
    Represents element of Clusterization. Groups together elements.
    """

    def __init__(self, cluster_id: int, elements: List[T], nodes_orders: List[int]) -> None:
        self.id = cluster_id
        self.elements = list(elements)
        self.importance = max(nodes_orders)

    def clean_cluster(self) -> None:
        self.elements = []
        self.importance = 0

    def add_elements(self, elements: List[T], importance: int) -> None:
        self.elements.extend(elements)
        self.importance = max(self.importance, importance)


class Clusterization(Generic[T]):
    """
    Handles group of node clusters allowing to add new cluster,
    delete existing one or merge existing clusters.
    """

    def __init__(self, id_fn: Optional[Callable[[T], Hashable]] = None) -> None:
        self.clusters: Dict[int, Cluster[T]] = {}
        self._element_to_cluster: Dict[Hashable, int] = {}
        if id_fn is None:
            self._id_fn = lambda x: x.id
        else:
            self._id_fn = id_fn

    def get_cluster_by_id(self, cluster_id: int) -> Cluster[T]:
        """
        Returns cluster according to provided cluster_id.

        :param cluster_id: Id of the cluster.
        :return: Cluster according to provided `cluster_id`.
        """
        if cluster_id not in self.clusters:
            raise IndexError("No cluster with id = {}".format(cluster_id))
        return self.clusters[cluster_id]

    def get_cluster_containing_element(self, element_id: Hashable) -> Cluster[T]:
        """
        Returns cluster containing element with provided `element_id`.

        :param element_id: Id of the element which is in cluster.
        :return: Cluster containing element with provided `element_id`.
        """
        if element_id not in self._element_to_cluster:
            raise IndexError("No cluster for node with id = {}".format(element_id))
        return self.get_cluster_by_id(self._element_to_cluster[element_id])

    def is_node_in_clusterization(self, node_id: int) -> bool:
        """
        Returns whether node with provided `node_id` is in clusterization.

        :param node_id: Id of the node to test.
        :return: Whether node with provided `node_id` is in clusterization.
        """
        return node_id in self._element_to_cluster

    def add_cluster(self, cluster: Cluster[T]) -> None:
        """
        Adds provided cluster to clusterization.

        :param cluster: Cluster to add.
        """
        cluster_id = cluster.id
        if cluster_id in self.clusters:
            raise IndexError("Cluster with index = {} already exist".format(cluster_id))
        self.clusters[cluster_id] = cluster
        for elt in cluster.elements:
            self._element_to_cluster[self._id_fn(elt)] = cluster_id  # type: ignore[no-untyped-call]

    def delete_cluster(self, cluster_id: int) -> None:
        """
        Removes cluster with `cluster_id` from clusterization.

        :param cluster_id: Id of a cluster to delete.
        """
        if cluster_id not in self.clusters:
            raise IndexError("No cluster with index = {} to delete".format(cluster_id))
        for elt in self.clusters[cluster_id].elements:
            node_id = self._id_fn(elt)  # type: ignore[no-untyped-call]
            self._element_to_cluster.pop(node_id)
        self.clusters.pop(cluster_id)

    def get_all_clusters(self) -> List[Cluster[T]]:
        """
        Returns list of all clusters in clusterization.

        :return: List of all clusters in clusterization.
        """
        return list(self.clusters.values())

    def get_all_nodes(self) -> List[T]:
        """
        Returns list all elements of all clusters in clusterization.

        :return: List all elements of all clusters in clusterization.
        """
        all_elements = []
        for cluster in self.clusters.values():
            all_elements.extend(cluster.elements)
        return all_elements

    def merge_clusters(self, first_id: int, second_id: int) -> None:
        """
        Merges two clusters with provided ids.

        :param first_id: Id of the first cluster to merge.
        :param second_id: Id of the second cluster to merge.
        """
        cluster_1 = self.get_cluster_by_id(first_id)
        cluster_2 = self.get_cluster_by_id(second_id)
        if cluster_1.importance > cluster_2.importance:
            cluster_1.add_elements(cluster_2.elements, cluster_2.importance)
            for elt in cluster_2.elements:
                self._element_to_cluster[self._id_fn(elt)] = first_id  # type: ignore[no-untyped-call]
            self.clusters.pop(second_id)
        else:
            cluster_2.add_elements(cluster_1.elements, cluster_1.importance)
            for elt in cluster_1.elements:
                self._element_to_cluster[self._id_fn(elt)] = second_id  # type: ignore[no-untyped-call]
            self.clusters.pop(first_id)

    def merge_list_of_clusters(self, clusters: List[int]) -> None:
        """
        Merges provided clusters.

        :param clusters: List of clusters to merge.
        """
        clusters = list(set(clusters))
        clusters.sort(key=lambda x: self.get_cluster_by_id(x).importance)
        for cluster_id in clusters[:-1]:
            self.merge_clusters(clusters[-1], cluster_id)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from nncf.tensorflow.graph.metatypes.common import ALL_LAYER_METATYPES_WITH_WEIGHTS
from nncf.tensorflow.graph.metatypes.common import GENERAL_CONV_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.common import NORMALIZATION_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.matcher import get_keras_layer_metatype
from nncf.tensorflow.graph.utils import unwrap_layer
from nncf.tensorflow.layers.operation import InputType


def get_channel_size(input_shape, channel_axes):
    if not isinstance(channel_axes, (list, tuple)):
        channel_axes = [channel_axes]
    size = 1
    for axis in channel_axes:
        size *= input_shape[axis]
    return size


def get_channel_axis(input_type, input_name, layer):
    if input_type == InputType.INPUTS:
        return get_input_channel_axis(layer)
    return get_weight_channel_axis(layer, input_name)


def get_data_format(layer):
    return getattr(layer, "data_format", "channels_last")


def get_input_channel_axis(layer):
    original_layer = unwrap_layer(layer)
    layer_metatype = get_keras_layer_metatype(original_layer, determine_subtype=False)
    data_format = get_data_format(original_layer)
    if layer_metatype in GENERAL_CONV_LAYER_METATYPES:
        return -1 if data_format == "channels_last" else -1 - original_layer.rank
    if layer_metatype in NORMALIZATION_LAYER_METATYPES:
        return original_layer.axis

    return -1 if data_format == "channels_last" else 1


def get_weight_channel_axis(layer, weight_attr):
    original_layer = unwrap_layer(layer)
    layer_metatype = get_keras_layer_metatype(original_layer, determine_subtype=False)
    if layer_metatype in ALL_LAYER_METATYPES_WITH_WEIGHTS:
        for weight_def in layer_metatype.weight_definitions:
            if weight_def.weight_attr_name == weight_attr:
                return weight_def.channel_axes

    return -1


def get_weight_shape(layer: tf.keras.layers.Layer, weight_attr: str) -> tf.TensorShape:
    original_layer = unwrap_layer(layer)
    weight = getattr(original_layer, weight_attr)
    return weight.shape



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any, Dict, Tuple, Union, cast

import networkx as nx  # type: ignore[import-untyped]
import torch
from torch import nn

from nncf.common.logging import nncf_logger as logger
from nncf.experimental.torch2.function_hook.graph.graph_utils import ConstMeta
from nncf.experimental.torch2.function_hook.graph.graph_utils import EdgeMeta
from nncf.experimental.torch2.function_hook.graph.graph_utils import FunctionMeta
from nncf.experimental.torch2.function_hook.graph.graph_utils import InOutMeta
from nncf.experimental.torch2.function_hook.graph.graph_utils import NodeType
from nncf.experimental.torch2.function_hook.graph.graph_utils import TensorInfo
from nncf.experimental.torch2.function_hook.graph.graph_utils import TensorMeta
from nncf.experimental.torch2.function_hook.graph.graph_utils import TensorSource
from nncf.experimental.torch2.function_hook.hook_executor_mode import FunctionHookMode
from nncf.experimental.torch2.function_hook.hook_executor_mode import OpMeta
from nncf.experimental.torch2.function_hook.hook_storage import HookStorage
from nncf.experimental.torch2.function_hook.weak_map import WeakUnhashableKeyMap
from nncf.experimental.torch2.function_hook.wrapper import ForwardWithHooks
from nncf.experimental.torch2.function_hook.wrapper import get_hook_storage


class GraphBuilderMode(FunctionHookMode):
    """
    A class that extends the `FunctionHookMode` and constructs nx.MultiDiGraph representing
    the operations within a PyTorch model. Each operation in the model is represented as a node in the graph.

    :param next_node_id: The index of the last added node in the graph.
    :param graph: An instance of `networkx.MultiDiGraph` used to represent the graph of model.
    :param tensor_info: A dictionary mapping tensor objects to tensor info.
    """

    def __init__(self, model: nn.Module, hook_storage: HookStorage):
        """
        Initialize the GraphBuilderMode.

        :param model: The PyTorch model to which the hooks will be applied.
        :param hook_storage: Storage for hooks to be executed.
        """
        super().__init__(model=model, hook_storage=hook_storage)
        self.next_node_id: int = 0
        self.graph: nx.MultiDiGraph = nx.MultiDiGraph()
        self.tensor_info: WeakUnhashableKeyMap[Union[torch.Tensor, torch.nn.Parameter], TensorInfo] = (
            WeakUnhashableKeyMap()
        )

        for name, parameter in self.model.named_parameters():
            self.tensor_info[parameter] = TensorInfo(
                tensor_source=TensorSource.parameter,
                shape=tuple(parameter.shape),
                dtype=parameter.dtype,
                output_port_id=0,
                source_node_id=None,
                name_in_model=name,
            )
        for name, buffer in self.model.named_buffers():
            self.tensor_info[buffer] = TensorInfo(
                tensor_source=TensorSource.buffer,
                shape=tuple(buffer.shape),
                dtype=buffer.dtype,
                output_port_id=0,
                source_node_id=None,
                name_in_model=name,
            )

    def __enter__(self) -> GraphBuilderMode:
        """
        Overload __enter__ to get correct return type hint.
        """
        super().__enter__()
        return self

    def register_new_node_id(self) -> int:
        """
        Generates and returns a unique id for a new node in the graph.

        :return: A unique id for a new node.
        """
        key = self.next_node_id
        self.next_node_id += 1
        return key

    def register_node_for_model_input_tensor(self, node_name: str, tensor: torch.Tensor) -> None:
        """
        Register a node for model input.

        :param node_name: The name of the input node to be registered.
        :param tensor: The tensor associated with the model's input.
        """
        node_id = self.register_new_node_id()
        with self.disable():
            self.tensor_info[tensor] = TensorInfo(
                tensor_source=TensorSource.input,
                shape=tuple(tensor.shape),
                dtype=tensor.dtype,
                output_port_id=0,
                source_node_id=node_id,
                name_in_model=None,
            )
            self.graph.add_node(node_id, type=NodeType.input, meta=InOutMeta.from_tensor(tensor, node_name))
        logger.debug(f"GraphBuilderMode.register_node_for_model_input_tensor: {node_id=} {node_name=}")

    def execute_hooks_for_model_input(self, name: str, value: Any) -> Any:
        """
        Overloaded function to register a model input to the graph.

        :param name: The name of the input argument.
        :param value: The value of the input argument.
        :return: The processed value after the hook is executed.
        """
        if isinstance(value, torch.Tensor):
            self.register_node_for_model_input_tensor(name, value)
        return super().execute_hooks_for_model_input(name, value)

    def register_node_for_model_output_tensor(self, node_name: str, tensor: torch.Tensor) -> None:
        """
        Registers a node for a model's output tensor in graph, creating an edge between
        the tensor's source node and the output node.

        :param node_name: The name of the output node to be registered.
        :param tensor: The tensor associated with the model's output.
        """
        node_id = self.register_new_node_id()
        tensor_info = self.tensor_info.get(tensor)
        if tensor_info is not None and isinstance(tensor_info.output_port_id, int):
            with self.disable():
                self.graph.add_edge(
                    tensor_info.source_node_id,
                    node_id,
                    meta=EdgeMeta.from_tensor(tensor, input_port=0, output_port=tensor_info.output_port_id),
                )
                self.graph.add_node(node_id, type=NodeType.output, meta=InOutMeta.from_tensor(tensor, node_name))
        logger.debug(f"GraphBuilderMode.register_node_for_model_input_tensor: {node_id=} {node_name=}")

    def execute_hooks_for_model_output(self, name: str, value: Any) -> Any:
        """
        Overloaded function to register a model output to the graph.

        :param name: The name of the input argument.
        :param value: The value of the input argument.
        :return: The processed value after the hook is executed.
        """
        value = super().execute_hooks_for_model_output(name, value)
        if isinstance(value, torch.Tensor):
            self.register_node_for_model_output_tensor(name, value)
        return value

    def execute_pre_hooks(
        self, args: Tuple[Any, ...], kwargs: Dict[str, Any], op_meta: OpMeta
    ) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        """
        Overloaded function to register a model output to the graph.

        :param args: The arguments to the function.
        :param kwargs: The keyword arguments to the function.
        :param op_meta: Metadata for the operation.
        :return: The modified arguments and keyword arguments after pre-hooks.
        """
        _args, _kwargs = super().execute_pre_hooks(args, kwargs, op_meta)
        self.register_op_node(_args, _kwargs, op_meta)
        return _args, _kwargs

    def process_tensor_attributes(self, output: torch.Tensor, op_meta: OpMeta) -> None:
        """
        Processes special functions like attribute access on tensors using grad_fn.

        Process attributes: detect .T and .mT attributes of tensor
            .T - permute((1, 0))
            .mT - transpose(-2, -1)

        :param output: The output tensor.
        :param op_meta: Metadata about the operation.
        """
        fn_name = None
        fn_kwargs = None

        if output.grad_fn is not None:
            if output.grad_fn.name() == "TransposeBackward0":
                fn_name = "transpose"
                # grad_fn collect arguments as _saved_dim0=18446744073709551614
                # Use static arguments for .mT
                # https://pytorch.org/docs/stable/tensors.html#torch.Tensor.mT
                fn_kwargs = {
                    "dim0": -2,
                    "dim1": -1,
                }
            if output.grad_fn.name() == "PermuteBackward0":
                fn_name = "permute"
                fn_kwargs = {"dims": output.grad_fn._saved_dims}  # type: ignore[attr-defined]

        if fn_name is not None and fn_kwargs is not None:
            self.graph.nodes[op_meta.extra_info["node_id"]]["meta"].fn_name = fn_name
            self.graph.nodes[op_meta.extra_info["node_id"]]["meta"].kwargs = fn_kwargs

    def execute_post_hooks(self, outputs: Any, op_meta: OpMeta) -> Any:
        """
        Overload execute_post_hooks to correct registered node for operation.
        Process __get__ function, to detect permute and transpose operation
        and remove node if operation return not tensor.
        """
        if op_meta.func.__name__ == "__get__":
            if isinstance(outputs, torch.Tensor):
                self.process_tensor_attributes(outputs, op_meta)
            else:
                # Remove the node corresponding to this operation from the graph, as non-tensor
                # outputs (like `tensor.shape` or similar) are not relevant for further algorithmic use.
                self.graph.remove_node(op_meta.extra_info["node_id"])
        outputs = super().execute_post_hooks(outputs, op_meta)
        return outputs

    def execute_hooks_for_parameter(self, value: torch.Tensor) -> torch.Tensor:
        """
        Overload execute_hooks_for_parameter to register parameters to the graph.

        :param value: The tensor to which the post-hook will be applied.
        :return: The processed tensor with the applied post-hook, if applicable.
        """
        tensor_info = self.tensor_info.get(value)
        if (
            tensor_info is not None
            and tensor_info.source_node_id is None
            and tensor_info.tensor_source in [TensorSource.buffer, TensorSource.parameter]
            and isinstance(tensor_info.name_in_model, str)
        ):
            node_id = self.register_new_node_id()
            with self.disable():
                self.graph.add_node(
                    node_id,
                    type=NodeType.const,
                    meta=ConstMeta.from_tensor(tensor=value, name_in_model=tensor_info.name_in_model),
                )
            tensor_info.source_node_id = node_id
            logger.debug(f"GraphBuilderMode._maybe_add_node_for_parameters: {node_id=} {tensor_info.name_in_model=}")
        return super().execute_hooks_for_parameter(value)

    def register_op_input_tensor(self, tensor: torch.Tensor, op_node_id: int, port_id: int, op_meta: OpMeta) -> None:
        """
        Registers a tensor input by creating an edge between the tensor's source node and the current operation node.

        :param tensor: Input tensor for the operation.
        :param node_id: Id if operation node.
        :param port_id: Port id of input argument.
        :param op_meta: Metadata about the operation.
        """
        tensor_info = self.tensor_info.get(tensor, None)
        if tensor_info is None:
            logger.debug(f"GraphBuilderMode.register_op_input_tensor: unknown tensor {op_meta.op_name=} {port_id=}")
            return
        self.graph.add_edge(
            tensor_info.source_node_id,
            op_node_id,
            meta=EdgeMeta.from_tensor(tensor, input_port=port_id, output_port=tensor_info.output_port_id),
        )

    def register_op_input(self, arg: Any, node_id: int, port_id: int, op_meta: OpMeta) -> Any:
        """
        Registers an operation input. Handles tensors or collections of tensors on one input port.

        :param arg: Operation input argument.
        :param node_id: Id if operation node.
        :param port_id: Port id of input argument.
        :param op_meta: Metadata about the operation.
        :return: Descriptor of the input. For a Tensor, this is a `TensorMeta` object.
             For a collection of Tensors, a collection of `TensorMeta` objects is returned.
             For other types, the original input `arg` is returned as-is.
        """
        if isinstance(arg, torch.Tensor):
            self.register_op_input_tensor(arg, node_id, port_id, op_meta)
            return TensorMeta.from_tensor(arg)
        elif isinstance(arg, (list, tuple, set)):
            op_attr = []
            for x in arg:
                if isinstance(x, torch.Tensor):
                    self.register_op_input_tensor(x, node_id, port_id, op_meta)
                    op_attr.append(TensorMeta.from_tensor(x))
                else:
                    op_attr.append(x)
            return op_attr
        return arg

    def register_op_node(self, args: Tuple[Any], kwargs: Dict[str, Any], op_meta: OpMeta) -> None:
        """
        Registers a new operation node in the graph.

        This method is responsible for:
            - adding a new node in the graph for the given operation and storing metadata such as the node ID,
              operation name, input arguments and keyword arguments;
            - adding edge associated with registered tensor.

        :param args: Positional arguments for the operation.
        :param kwargs: Keyword arguments for the operation.
        :param op_meta: Metadata about the operation, including the function being invoked.
        """
        node_id = self.register_new_node_id()
        op_meta.extra_info["node_id"] = node_id
        op_name = op_meta.op_name

        op_attrs = []
        op_kwargs = {}
        for port_id, arg in enumerate(args):
            op_attr = self.register_op_input(arg, node_id, port_id, op_meta)
            op_attrs.append(op_attr)

        for port_id, (name, arg) in enumerate(kwargs.items(), start=len(args)):
            op_attr = self.register_op_input(arg, node_id, port_id, op_meta)
            op_kwargs[name] = op_attr

        self.graph.add_node(
            node_id,
            type=NodeType.fn_call,
            meta=FunctionMeta(op_name=op_name, fn_name=op_meta.func.__name__, args=tuple(op_attrs), kwargs=op_kwargs),
        )

        logger.debug(f"GraphBuilderMode.process_op_inputs: {node_id=} {op_name=} {op_attrs=} {op_kwargs=}")

    def process_post_function_hooks_for_value(self, value: Any, op_meta: OpMeta, port_id: int) -> Any:
        """
        Overload process_post_function_hooks_for_value to get register output tensors to the graph

        :param output: The output of the function.
        :param op_meta: Metadata for the operation.
        :return: The modified output after post-hooks.
        """
        if isinstance(value, torch.Tensor):
            with self.disable():
                self.tensor_info[value] = TensorInfo(
                    tensor_source=TensorSource.function,
                    shape=tuple(value.shape),
                    dtype=value.dtype,
                    output_port_id=port_id,
                    source_node_id=op_meta.extra_info["node_id"],
                    name_in_model=None,
                )
        return super().process_post_function_hooks_for_value(value, op_meta, port_id)


def build_graph(model: nn.Module, *args: Any, **kwargs: Any) -> nx.MultiDiGraph:
    """
    Constructs a computational graph of the given model.

    This function builds a directed graph `nx.MultiDiGraph` representing the operations
    and data flow within the model by leveraging hooks by using GraphBuilderMode.

    :param model: The PyTorch model for which the computational graph will be built.
    :return: A nx.MultiDiGraph where nodes represent operations of model.
    """

    with torch.enable_grad():  # type: ignore
        # Gradient use to get information about __get__ functions to detect tensor.(T, mT) attributes
        with GraphBuilderMode(model=model, hook_storage=get_hook_storage(model)) as ctx:
            args, kwargs = ctx.process_model_inputs(args, kwargs)
            wrapped_forward = cast(ForwardWithHooks, model.forward)
            outputs = wrapped_forward._func(*args, **kwargs)
            outputs = ctx.process_model_outputs(outputs)
    return ctx.graph



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from nncf.common.graph.transformations.commands import TransformationCommand


class TransformationLayout:
    """
    Represents a list of transformation commands that has been prepared to be
    correctly applied to a model. In practice, stacking compression algorithms or
    a specific model transformation algorithm imposes some restrictions on the
    order in which transformation commands are applied. `TransformationLayout`
    addresses these issues.
    """

    def __init__(self) -> None:
        """
        Initialize Transformation Layout.
        """
        self._transformations: List[TransformationCommand] = []

    @property
    def transformations(self) -> List[TransformationCommand]:
        return self._transformations

    def register(self, transformation: TransformationCommand) -> None:
        """
        Registers the transformation command in the transformation layout.

        :param transformation: The transformation command to be registered in
            the transformation layout.
        """
        self.transformations.append(transformation)

    def update(self, other: "TransformationLayout") -> None:
        """
        D.update(other), updates D from other.

        :param other: Another transformation layout.
        """
        for transformation in other.transformations:
            self.register(transformation)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Set, Tuple

import numpy as np

import nncf
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.hardware.config import HWConfig
from nncf.common.quantization.structs import QuantizerConfig
from nncf.experimental.common.tensor_statistics.collectors import REDUCERS_MAP
from nncf.experimental.common.tensor_statistics.collectors import TensorReducerBase
from nncf.onnx.graph.metatypes import onnx_metatypes as om
from nncf.onnx.graph.metatypes.groups import ELEMENTWISE_OPERATIONS
from nncf.onnx.graph.metatypes.groups import MATMUL_METATYPES
from nncf.onnx.graph.node_utils import get_input_edges_mapping
from nncf.onnx.graph.node_utils import get_quantized_tensor_shape
from nncf.onnx.graph.node_utils import get_weight_quantization_axis
from nncf.onnx.graph.transformations.commands import ONNXQuantizerInsertionCommand
from nncf.onnx.graph.transformations.commands import ONNXTargetPoint
from nncf.onnx.hardware.config import ONNXHWConfig
from nncf.onnx.quantization.default_quantization import DEFAULT_ONNX_QUANT_TRAIT_TO_OP_DICT
from nncf.onnx.quantization.quantizer_parameters import convert_fq_params_to_onnx_params
from nncf.parameters import ModelType
from nncf.parameters import TargetDevice
from nncf.quantization.algorithms.min_max.backend import MinMaxAlgoBackend
from nncf.quantization.fake_quantize import FakeConvertParameters
from nncf.quantization.fake_quantize import FakeQuantizeParameters
from nncf.quantization.range_estimator import StatisticsType


class ONNXMinMaxAlgoBackend(MinMaxAlgoBackend):
    @property
    def preserved_metatypes(self) -> List[OperatorMetatype]:
        return []

    @property
    def mat_mul_metatypes(self) -> List[OperatorMetatype]:
        return MATMUL_METATYPES

    @property
    def post_processing_metatypes(self) -> List[OperatorMetatype]:
        return [om.ONNXTopKMetatype, om.ONNXNonMaxSuppressionMetatype]

    @property
    def conv_metatypes(self) -> List[OperatorMetatype]:
        return [om.ONNXConvolutionMetatype]

    @property
    def elementwise_metatypes(self) -> List[OperatorMetatype]:
        return ELEMENTWISE_OPERATIONS

    @property
    def overflow_fix_metatypes(self) -> List[OperatorMetatype]:
        return [
            om.ONNXConvolutionMetatype,
            om.ONNXConvolutionTransposeMetatype,
            *MATMUL_METATYPES,
        ]

    @property
    def add_metatypes(self) -> List[OperatorMetatype]:
        return [om.ONNXAddLayerMetatype]

    @property
    def group_conv_metatypes(self) -> List[OperatorMetatype]:
        return self.conv_metatypes

    @property
    def shapeof_metatypes(self) -> List[OperatorMetatype]:
        return [om.ONNXShapeMetatype]

    @property
    def dropout_metatypes(self) -> List[OperatorMetatype]:
        return []

    @property
    def read_variable_metatypes(self) -> List[OperatorMetatype]:
        return []

    @property
    def scaled_dot_product_attention_metatypes(self) -> List[OperatorMetatype]:
        return []

    @property
    def scales_unification_map(self) -> Dict[OperatorMetatype, OperatorMetatype]:
        return {om.ONNXConcatMetatype: self.overflow_fix_metatypes + self.scaled_dot_product_attention_metatypes}

    @property
    def hw_config(self) -> HWConfig:
        return ONNXHWConfig

    @property
    def quant_trait_op_dict(self) -> Dict[int, OperatorMetatype]:
        return DEFAULT_ONNX_QUANT_TRAIT_TO_OP_DICT

    @property
    def reducer_map(self) -> Dict[StatisticsType, TensorReducerBase]:
        return REDUCERS_MAP

    @property
    def supports_inplace_statistics(self) -> bool:
        return False

    @staticmethod
    def get_start_nodes_for_activation_path_tracing(
        nncf_graph: NNCFGraph,
    ) -> List[NNCFNode]:
        return nncf_graph.get_input_nodes()

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> ONNXTargetPoint:
        return ONNXTargetPoint(target_type, target_node_name, port_id)

    @staticmethod
    def create_quantizer_insertion_command(
        nncf_graph: NNCFGraph,
        target_point: ONNXTargetPoint,
        quantizer_config: QuantizerConfig,
        parameters: FakeQuantizeParameters,
    ) -> ONNXQuantizerInsertionCommand:
        tensor_type = np.int8 if np.any(parameters.input_low.data < 0) else np.uint8
        is_weight = target_point.is_weight_target_point()
        if is_weight:
            tensor_type = np.int8  # The weight is restricted to have only signed range
        nncf_input_node_next_nodes = ONNXMinMaxAlgoBackend._get_input_edges_mapping(nncf_graph)
        node = nncf_graph.get_node_by_name(target_point.target_node_name)
        axis = ()
        if quantizer_config.per_channel:
            axis = ONNXMinMaxAlgoBackend.get_weight_quantization_axes(node, target_point, None) if is_weight else (1,)
        onnx_parameters = convert_fq_params_to_onnx_params(parameters, quantizer_config.num_bits, tensor_type, axis)
        return ONNXQuantizerInsertionCommand(target_point, nncf_input_node_next_nodes, onnx_parameters)

    @staticmethod
    def create_unified_scales_quantizers_insertion_commands(
        nncf_graph: NNCFGraph,
        target_points: List[ONNXTargetPoint],
        quantizer_config: QuantizerConfig,
        parameters: FakeQuantizeParameters,
    ) -> List[ONNXQuantizerInsertionCommand]:
        return [
            ONNXMinMaxAlgoBackend.create_quantizer_insertion_command(
                nncf_graph, target_point, quantizer_config, parameters
            )
            for target_point in target_points
        ]

    @staticmethod
    def create_convert_insertion_command(
        target_point: ONNXTargetPoint,
        parameters: FakeConvertParameters,
    ) -> TransformationCommand:
        raise nncf.InternalError("FakeConvert insertion not implemented in ONNX backend!")

    @staticmethod
    def _get_input_edges_mapping(nncf_graph: NNCFGraph):
        return get_input_edges_mapping(nncf_graph)

    @staticmethod
    def get_target_point_shape(nncf_graph: NNCFGraph, node: NNCFNode, target_point: ONNXTargetPoint) -> Tuple[int, ...]:
        return get_quantized_tensor_shape(nncf_graph, node, target_point)

    @staticmethod
    def get_weight_quantization_axes(node: NNCFNode, target_point: ONNXTargetPoint, ndims: int) -> Tuple[int]:
        return (get_weight_quantization_axis(node, target_point.port_id),)

    @staticmethod
    def get_weight_tensor_port_ids(node: NNCFNode, graph: NNCFGraph) -> List[Optional[int]]:
        return list(node.layer_attributes.weight_attrs.keys())

    @staticmethod
    def get_ignored_metatypes(model_type: ModelType, device: TargetDevice) -> List[OperatorMetatype]:
        types = []
        if model_type == ModelType.TRANSFORMER:
            types = [
                om.ONNXAddLayerMetatype,
                om.ONNXPowMetatype,
                om.ONNXSqueezeMetatype,
                om.ONNXSubMetatype,
                om.ONNXAveragePoolMetatype,
                om.ONNXGlobalAveragePoolMetatype,
                om.ONNXReduceMeanMetatype,
                om.ONNXReduceL2Metatype,
                om.ONNXReduceSumMetatype,
                om.ONNXDivLayerMetatype,
                om.ONNXMaximumMetatype,
                om.ONNXSqrtMetatype,
                om.ONNXReciprocalMetatype,
                om.ONNXBatchNormMetatype,
                # Сomparison operations
                om.ONNXGreaterMetatype,
                om.ONNXLessMetatype,
                om.ONNXEqualMetatype,
            ]
            if device != TargetDevice.CPU_SPR:
                types.append(om.ONNXMulLayerMetatype)
        return types

    @staticmethod
    def get_ignored_names_by_layer_attributes(nncf_graph: NNCFGraph) -> Set[str]:
        return set()

    def get_weight_nodes(self, nncf_graph: NNCFGraph) -> List[NNCFNode]:
        return [node for node in nncf_graph.get_all_nodes() if node.layer_attributes.has_weight()]

    @staticmethod
    def get_weight_name(nncf_graph: NNCFGraph, target_point: ONNXTargetPoint) -> str:
        node_name, port_id = target_point.target_node_name, target_point.port_id
        return nncf_graph.get_node_by_name(node_name).layer_attributes.weight_attrs[port_id]["name"]

    @staticmethod
    def should_quantize_weight(weight_name: str, quantized_weight_names: Set[str]) -> bool:
        # If the nodes share one weight tensor, we should have only one quantizer on that
        return weight_name not in quantized_weight_names

    def is_matmul_with_constant(self, node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return node.metatype in self.mat_mul_metatypes and node.layer_attributes.has_weight()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import TypeVar

import torch.nn

import nncf
from nncf import NNCFConfig
from nncf.common.composite_compression import CompositeCompressionAlgorithmBuilder
from nncf.common.composite_compression import CompositeCompressionAlgorithmController
from nncf.common.composite_compression import CompositeCompressionLoss
from nncf.config.extractors import extract_algorithm_names
from nncf.torch.algo_selector import PT_COMPRESSION_ALGORITHMS
from nncf.torch.compression_method_api import PTCompressionAlgorithmBuilder
from nncf.torch.compression_method_api import PTCompressionAlgorithmController
from nncf.torch.compression_method_api import PTCompressionLoss
from nncf.torch.graph.transformations.layout import PTTransformationLayout
from nncf.torch.model_transformer import PTModelTransformer
from nncf.torch.nncf_network import NNCFNetwork

TModel = TypeVar("TModel")


class PTCompositeCompressionLoss(CompositeCompressionLoss, PTCompressionLoss):
    def __init__(self):
        super().__init__()
        self._child_losses = torch.nn.ModuleList()

    @property
    def child_losses(self) -> torch.nn.ModuleList:
        return self._child_losses


class PTCompositeCompressionAlgorithmBuilder(CompositeCompressionAlgorithmBuilder, PTCompressionAlgorithmBuilder):
    def __init__(self, config: NNCFConfig, should_init: bool = True):
        super().__init__(config, should_init)

        algo_names = extract_algorithm_names(config)
        if len(algo_names) < 2:
            raise nncf.ValidationError(
                "Composite algorithm builder must be supplied with a config with more than one "
                "compression algo specified!"
            )
        for algo_name in algo_names:
            algo_builder = PT_COMPRESSION_ALGORITHMS.get(algo_name)
            self._child_builders.append(algo_builder(config, should_init=should_init))

    def __bool__(self):
        return bool(self.child_builders)

    def apply_to(self, model: NNCFNetwork) -> NNCFNetwork:
        transformer = PTModelTransformer(model)
        layout = self.get_transformation_layout(model)
        transformed_model = transformer.transform(layout)

        self.initialize(transformed_model)

        return transformed_model

    def _build_controller(self, model: TModel) -> PTCompressionAlgorithmController:
        """
        Simple implementation of building controller without setting builder state and loading controller's one.
        Builds `PTCompositeCompressionAlgorithmController` to handle the additional
        modules, parameters, and hooks inserted into the model to enable
        algorithm-specific compression.

        :param model: The model with additional modifications necessary to enable
         algorithm-specific compression during fine-tuning.
        :return: The instance of the `PTCompositeCompressionAlgorithmController`.
        """
        if len(self._child_builders) == 1:
            return self._child_builders[0].build_controller(model)
        composite_ctrl = PTCompositeCompressionAlgorithmController(model)
        for builder in self.child_builders:
            composite_ctrl.add(builder.build_controller(model))
        return composite_ctrl

    def get_transformation_layout(self, model: TModel) -> PTTransformationLayout:
        """
        Computes necessary model transformations to enable algorithm-specific
        compression.

        :param model: The original uncompressed model.
        :return: The instance of the `PTTransformationLayout` class containing
            a list of algorithm-specific modifications.
        """
        transformations = PTTransformationLayout()
        for builder in self.child_builders:
            transformations.update(builder.get_transformation_layout(model))
        return transformations

    def initialize(self, model: TModel) -> None:
        for builder in self.child_builders:
            if builder.should_init:
                builder.initialize(model)

    def _get_transformation_layout(self, target_model: NNCFNetwork) -> PTTransformationLayout:
        pass  # Higher-level get_transformation_layout is overridden, no need to define this


class PTCompositeCompressionAlgorithmController(
    CompositeCompressionAlgorithmController, PTCompressionAlgorithmController
):
    def __init__(self, target_model: TModel):
        super().__init__(target_model)
        self._loss = PTCompositeCompressionLoss()

    def distributed(self):
        for ctrl in self.child_ctrls:
            ctrl.distributed()

    def prepare_for_export(self):
        for child_ctrl in self.child_ctrls:
            child_ctrl.prepare_for_export()

    @property
    def compression_rate(self) -> float:
        sum_compression_rate = 0
        not_none_compression_rate_cnt = 0
        for child_ctrl in self.child_ctrls:
            compression_rate = child_ctrl.compression_rate
            if compression_rate is not None:
                sum_compression_rate += sum_compression_rate
                not_none_compression_rate_cnt += 1
        return sum_compression_rate / max(not_none_compression_rate_cnt, 1)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict, List, Set, Tuple

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph import NNCFNodeName
from nncf.common.pruning.clusterization import Cluster
from nncf.common.pruning.clusterization import Clusterization
from nncf.common.pruning.mask_propagation import MaskPropagationAlgorithm
from nncf.common.pruning.structs import PrunedLayerInfoBase
from nncf.common.pruning.symbolic_mask import SymbolicMask
from nncf.common.pruning.symbolic_mask import SymbolicMaskProcessor
from nncf.common.pruning.utils import PruningOperationsMetatypeRegistry
from nncf.common.pruning.utils import get_input_masks
from nncf.common.pruning.utils import get_next_nodes_of_types
from nncf.common.pruning.utils import get_output_channels
from nncf.common.pruning.utils import get_prunable_layers_in_out_channels
from nncf.common.pruning.utils import get_rounded_pruned_element_number


class ShapePruningProcessor:
    """
    Collection of shape pruning functions. Class instance keeps
    only parameters that are constant during
    compression algorithms execution.
    """

    def __init__(self, prunable_types: List[str], pruning_operations_metatype: PruningOperationsMetatypeRegistry):
        """
        Constructor.

        :param prunable_types: Types of nodes that will be returned.
        :param pruning_operations_metatype: Metatypes of nodes that will be returned.
        """
        self._prunable_types = prunable_types
        self._pruning_operations_metatype = pruning_operations_metatype

    def calculate_in_out_channels_by_masks(
        self,
        graph: NNCFGraph,
        pruning_groups: Clusterization[PrunedLayerInfoBase],
        pruning_groups_next_nodes: Dict[int, List[Dict[str, Any]]],
        num_of_sparse_elements_by_node: Dict[NNCFNodeName, int],
    ) -> Tuple[Dict[str, int], Dict[str, int]]:
        """
        Imitates filters pruning by removing output filters zeroed by pruning masks in each pruning group
        and updating corresponding input channels number in `pruning_groups_next_nodes` nodes.

        :param graph: NNCFGraph.
        :param pruning_groups: A list of pruning groups.
        :param pruning_groups_next_nodes: A dictionary of next nodes of each pruning group.
        :param num_of_sparse_elements_by_node: A dictionary of num_of_sparse_elements of each pruning node.
        :return Dictionary of new input channels number {node_name: channels_num}
        """

        def get_sparser(_: Any) -> Callable[[NNCFNodeName], int]:
            def get_num_of_sparse_elements_by_node(node_name: NNCFNodeName) -> int:
                return num_of_sparse_elements_by_node[node_name]

            return get_num_of_sparse_elements_by_node

        return self._calculate_in_out_channels(get_sparser, graph, pruning_groups, pruning_groups_next_nodes)

    def calculate_in_out_channels_in_uniformly_pruned_model(
        self,
        graph: NNCFGraph,
        pruning_groups: Clusterization[PrunedLayerInfoBase],
        pruning_groups_next_nodes: Dict[int, List[Dict[str, Any]]],
        pruning_level: float,
    ) -> Tuple[Dict[str, int], Dict[str, int]]:
        """
        Imitates filters pruning by removing `pruning_level` percent of output filters in each pruning group
        and updating corresponding input channels number in `pruning_groups_next_nodes` nodes.

        :param graph: NNCFGraph.
        :param pruning_groups: A list of pruning groups.
        :param pruning_groups_next_nodes: A dictionary of next nodes of each pruning group.
        :param pruning_level: Target pruning level.
        :return Tuple of dictionaries of new input and output channels number {node_name: channels_num}
        """

        def get_sparser(full_output_channels: Dict[NNCFNodeName, int]) -> Callable[[NNCFNodeName], int]:
            def get_num_of_sparse_elements_by_node(node_name: NNCFNodeName) -> int:
                old_out_channels = full_output_channels[node_name]
                return get_rounded_pruned_element_number(old_out_channels, pruning_level)

            return get_num_of_sparse_elements_by_node

        return self._calculate_in_out_channels(get_sparser, graph, pruning_groups, pruning_groups_next_nodes)

    def prune_cluster_shapes(
        self,
        cluster: Cluster[PrunedLayerInfoBase],
        pruned_elems: int,
        pruning_groups_next_nodes: Dict[int, List[Dict[str, Any]]],
        input_channels: Dict[NNCFNodeName, int],
        output_channels: Dict[NNCFNodeName, int],
    ) -> None:
        """
        Imitates filter pruning by removing `pruned_elems` elements from
        input/output channels corresponded to feeded cluster.

        :param cluster:  A PrunedLayerInfoBase cluster.
        :param pruned_elems: Amount of channels/elements to prune.
        :param pruning_groups_next_nodes: A dictionary of next nodes of each pruning group.
        :param input_channels: A dictionary of input channels number in prunable layers.
            Will be modified according to the filter pruning algorithm.
        :param output_channels: A dictionary of output channels number in prunable layers.
            Will be modified according to the filter pruning algorithm.
        """
        if not pruned_elems:
            return

        for node in cluster.elements:
            output_channels[node.node_name] -= pruned_elems
            if node.is_depthwise:
                input_channels[node.node_name] -= pruned_elems

        # Prune in channels in all next nodes
        next_nodes_info = pruning_groups_next_nodes[cluster.id]
        for next_node_info in next_nodes_info:
            input_channels[next_node_info["node_name"]] -= pruned_elems * next_node_info["sparse_multiplier"]

    def _calculate_in_out_channels(
        self,
        sparse_elements_counter_getter: Callable[[Dict[NNCFNodeName, int]], Callable[[NNCFNodeName], int]],
        graph: NNCFGraph,
        pruning_groups: Clusterization[PrunedLayerInfoBase],
        pruning_groups_next_nodes: Dict[int, List[Dict[str, Any]]],
    ) -> Tuple[Dict[str, int], Dict[str, int]]:
        full_input_channels, full_output_channels = get_prunable_layers_in_out_channels(graph)
        tmp_in_channels, tmp_out_channels = full_input_channels.copy(), full_output_channels.copy()
        sparse_elements_counter = sparse_elements_counter_getter(full_output_channels)

        for group in pruning_groups.get_all_clusters():
            layer_name = group.elements[0].node_name
            assert all(tmp_out_channels[layer_name] == tmp_out_channels[node.node_name] for node in group.elements)
            # Prune all nodes in cluster (by output channels)
            num_of_sparse_elems = sparse_elements_counter(layer_name)

            self.prune_cluster_shapes(
                cluster=group,
                pruned_elems=num_of_sparse_elems,
                input_channels=tmp_in_channels,
                output_channels=tmp_out_channels,
                pruning_groups_next_nodes=pruning_groups_next_nodes,
            )

        return tmp_in_channels, tmp_out_channels

    def _get_next_node_sparse_multiplier(
        self, graph: NNCFGraph, next_node: NNCFNode, cluster: Cluster[PrunedLayerInfoBase]
    ) -> int:
        cluster_nodes_idxs = {node.nncf_node_id for node in cluster.elements}
        for input_mask in get_input_masks(next_node, graph):
            if input_mask is None:
                continue
            for mask_producer in input_mask.mask_producers:
                if mask_producer.id in cluster_nodes_idxs:
                    return mask_producer.sparse_multiplier

        raise nncf.ValidationError(f"Next node for cluster {cluster.elements} doesn't have closing mask")

    def get_next_nodes(
        self, graph: NNCFGraph, pruning_groups: Clusterization[PrunedLayerInfoBase]
    ) -> Dict[int, List[Dict[str, Any]]]:
        """
        Finds nodes of `prunable_types` types that receive the output of a pruned cluster as input
        and collects all info specified in NextNode.

        :param graph: NNCFGraph.
        :param pruning_groups: `Clusterization` of pruning groups.
        :return Dictionary of next nodes by cluster {cluster_id: [node]}.
        """
        # 1. Propagate symbolic masks throught the net
        for pruned_layer_info in pruning_groups.get_all_nodes():
            node = graph.get_node_by_id(pruned_layer_info.nncf_node_id)
            node.attributes["output_mask"] = SymbolicMask(get_output_channels(node), node.node_id)

        MaskPropagationAlgorithm(graph, self._pruning_operations_metatype, SymbolicMaskProcessor).mask_propagation()

        # 2. Find next nodes and correspondent sparse multipliers
        next_nodes: Dict[int, List[Dict[str, Any]]] = {}
        for cluster in pruning_groups.get_all_clusters():
            next_nodes_cluster: Set[NNCFNode] = set()
            cluster_nodes: Set[NNCFNode] = set()
            for pruned_layer_info in cluster.elements:
                nncf_cluster_node = graph.get_node_by_id(pruned_layer_info.nncf_node_id)
                cluster_nodes.add(nncf_cluster_node)
                curr_next_nodes = get_next_nodes_of_types(graph, nncf_cluster_node, self._prunable_types)

                next_nodes_cluster = next_nodes_cluster.union(curr_next_nodes)
            next_nodes_cluster = next_nodes_cluster - cluster_nodes
            next_nodes[cluster.id] = []
            for next_node in next_nodes_cluster:
                sparse_multiplier = self._get_next_node_sparse_multiplier(graph, next_node, cluster)
                next_nodes[cluster.id].append(
                    {"node_name": next_node.node_name, "sparse_multiplier": sparse_multiplier}
                )

        # 3. Clean graph output shapes
        for node in graph.get_all_nodes():
            node.attributes["output_shape"] = None

        return next_nodes



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Tuple

import numpy as np
import openvino.runtime as ov

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.layer_attributes import ConvolutionLayerAttributes
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.tensor_statistics.collectors import TensorStatisticCollectorBase
from nncf.experimental.common.tensor_statistics.collectors import MedianAggregator
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.common.tensor_statistics.statistics import MinMaxTensorStatistic
from nncf.openvino.graph import node_utils
from nncf.openvino.graph.layout import OVLayoutElem
from nncf.openvino.graph.layout import get_conv_weights_layout_from_node
from nncf.openvino.graph.layout import get_linear_weights_layout_from_node
from nncf.openvino.graph.metatypes.groups import CONV_OPERATIONS
from nncf.openvino.graph.metatypes.openvino_metatypes import OVAddMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVConvolutionMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVDepthwiseConvolutionMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVGroupConvolutionMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVMatMulMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVSubtractMetatype
from nncf.openvino.graph.node_utils import create_bias_tensor
from nncf.openvino.graph.node_utils import get_bias_value
from nncf.openvino.graph.node_utils import get_weight_value
from nncf.openvino.graph.transformations.commands import OVTargetPoint
from nncf.openvino.statistics.collectors import OVQuantileReducer
from nncf.quantization.algorithms.channel_alignment.backend import ChannelAlignmentAlgoBackend
from nncf.quantization.algorithms.channel_alignment.backend import LayoutDescriptor


class OVChannelAlignmentAlgoBackend(ChannelAlignmentAlgoBackend):
    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> OVTargetPoint:
        return OVTargetPoint(target_type, target_node_name, port_id)

    @staticmethod
    def get_bias_value(node: NNCFNode, model: ov.Model, nncf_graph: NNCFGraph) -> np.ndarray:
        return get_bias_value(node, nncf_graph, model)

    @staticmethod
    def get_weight_value(node: NNCFNode, model: ov.Model, port_id: int) -> np.ndarray:
        return get_weight_value(node, model, port_id)

    @staticmethod
    def get_activation_port_ids_for_node(node: NNCFNode) -> Tuple[int, int]:
        return 0, 0

    @staticmethod
    def get_weights_port_ids_for_node(node: NNCFNode) -> Tuple[int, int]:
        return 0, 1

    @staticmethod
    def get_conv_metatypes():
        return [OVConvolutionMetatype, OVGroupConvolutionMetatype, OVDepthwiseConvolutionMetatype]

    @staticmethod
    def get_linear_metatypes():
        return [OVMatMulMetatype]

    @staticmethod
    def get_add_metatypes():
        return [OVAddMetatype, OVSubtractMetatype]

    @staticmethod
    def get_statistic_collector(
        reduction_axes, q: float, num_samples: int, inplace: bool
    ) -> TensorStatisticCollectorBase:
        tensor_collector = TensorCollector(MinMaxTensorStatistic)
        quantile_reducer = OVQuantileReducer(reduction_axes, (q, 1 - q), inplace)

        for port_id, container_key in enumerate([MinMaxTensorStatistic.MIN_STAT, MinMaxTensorStatistic.MAX_STAT]):
            aggregator = MedianAggregator(num_samples=num_samples, aggregation_axes=(0, 1))
            tensor_collector.register_statistic_branch(container_key, quantile_reducer, aggregator, port_id)
        return tensor_collector

    @staticmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return node_utils.is_node_with_bias(node, nncf_graph)

    @staticmethod
    def get_dims_descriptor(node: NNCFNode) -> LayoutDescriptor:
        if node.metatype in CONV_OPERATIONS:
            weights_layout = get_conv_weights_layout_from_node(node=node)
        elif node.metatype == OVMatMulMetatype:
            weights_layout = get_linear_weights_layout_from_node(node=node)
        else:
            raise nncf.InternalError(
                f"Metatype {node.metatype} of node {node.node_name} dimensions description retrieving is not supported"
            )

        if OVLayoutElem.GROUPS in weights_layout:
            # Using groups dim as output channels dim for ChannelAlignment algorithm
            # TODO(dlyakhov) support group convolutions with groups number not in [1, out_channels]
            return LayoutDescriptor(
                weights_layout.index(OVLayoutElem.GROUPS),
                weights_layout.index(OVLayoutElem.C_IN),
                node.metatype.output_channel_axis,
            )
        return LayoutDescriptor(
            weights_layout.index(OVLayoutElem.C_OUT) if OVLayoutElem.C_OUT in weights_layout else None,
            weights_layout.index(OVLayoutElem.C_IN),
            node.metatype.output_channel_axis,
        )

    @staticmethod
    def get_conv_layer_attributes(node: NNCFNode) -> ConvolutionLayerAttributes:
        return node.layer_attributes.layer_attributes

    @staticmethod
    def create_bias_tensor(node: NNCFNode, nncf_graph: NNCFGraph, value: Any) -> np.ndarray:
        return create_bias_tensor(node, nncf_graph, value)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Callable, Dict, List

import numpy as np

import nncf
from nncf.tensor import Tensor
from nncf.tensor.definitions import TensorBackend


def tensor_guard(func: callable):
    """
    A decorator that ensures that the first argument to the decorated function is a Tensor.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if isinstance(args[0], Tensor):
            return func(*args, **kwargs)
        raise NotImplementedError(f"Function `{func.__name__}` is not implemented for {type(args[0])}")

    return wrapper


def dispatch_list(fn: "functools._SingleDispatchCallable", tensor_list: List[Tensor], *args, **kwargs):
    """
    Dispatches the function to the type of the wrapped data of the first element in tensor_list.

    :param fn: A function wrapped by `functools.singledispatch`.
    :param tensor_list: List of Tensors.
    :return: The result value of the function call.
    """
    unwrapped_list = [i.data for i in tensor_list]
    return fn.dispatch(type(unwrapped_list[0]))(unwrapped_list, *args, **kwargs)


def dispatch_dict(fn: "functools._SingleDispatchCallable", tensor_dict: Dict[str, Tensor], *args, **kwargs):
    """
    Dispatches the function to the type of the wrapped data of the any element in tensor_dict.

    :param fn: A function wrapped by `functools.singledispatch`.
    :param tensor_dict: Dict of Tensors.
    :return: The result value of the function call.
    """
    unwrapped_dict = {}
    tensor_backend = None
    for key, tensor in tensor_dict.items():
        if tensor_backend is None:
            tensor_backend = type(tensor.data)
        else:
            if tensor_backend is not type(tensor.data):
                raise nncf.InternalError("All tensors in the dictionary should have the same backend")
        unwrapped_dict[key] = tensor.data

    return fn.dispatch(tensor_backend)(unwrapped_dict, *args, **kwargs)


def register_numpy_types(singledispatch_fn):
    """
    Decorator to register function to singledispatch for numpy classes.

    :param singledispatch_fn: singledispatch function.
    """

    def inner(func):
        singledispatch_fn.register(np.ndarray)(func)
        singledispatch_fn.register(np.generic)(func)
        singledispatch_fn.register(float)(func)
        return func

    return inner


def get_numeric_backend_fn(fn_name: str, backend: TensorBackend) -> Callable:
    """
    Returns a numeric function based on the provided function name and backend type.

    :param fn_name: The name of the numeric function.
    :param backend: The backend type for which the function is required.
    :return: The backend-specific numeric function.
    """
    if backend == TensorBackend.numpy:
        from nncf.tensor.functions import numpy_numeric

        return getattr(numpy_numeric, fn_name)
    if backend == TensorBackend.torch:
        from nncf.tensor.functions import torch_numeric

        return getattr(torch_numeric, fn_name)


def get_io_backend_fn(fn_name: str, backend: TensorBackend) -> Callable:
    """
    Returns a io function based on the provided function name and backend type.

    :param fn_name: The name of the numeric function.
    :param backend: The backend type for which the function is required.
    :return: The backend-specific io function.
    """
    if backend == TensorBackend.numpy:
        from nncf.tensor.functions import numpy_io

        return getattr(numpy_io, fn_name)
    if backend == TensorBackend.torch:
        from nncf.tensor.functions import torch_io

        return getattr(torch_io, fn_name)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, Dict, Set, Union

import torch

from nncf.api.compression import CompressionStage
from nncf.common.schedulers import StubCompressionScheduler
from nncf.common.statistics import NNCFStatistics
from nncf.common.tensor_statistics.collectors import ReductionAxes
from nncf.common.tensor_statistics.collectors import TensorStatisticCollectorBase
from nncf.config import NNCFConfig
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.tensor import Tensor
from nncf.torch.algo_selector import ZeroCompressionLoss
from nncf.torch.compression_method_api import PTCompressionAlgorithmBuilder
from nncf.torch.compression_method_api import PTCompressionAlgorithmController
from nncf.torch.dynamic_graph.context import no_nncf_trace
from nncf.torch.graph.transformations.commands import PTInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.graph.transformations.commands import TransformationPriority
from nncf.torch.graph.transformations.layout import PTTransformationLayout
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.return_types import maybe_get_values_from_torch_return_type


class TensorStatisticObservationPoint:
    def __init__(self, target_point: PTTargetPoint, reduction_shapes: Set[ReductionAxes] = None):
        self.target_point = target_point
        self.reduction_shapes = reduction_shapes

    def __hash__(self):
        return hash(self.target_point)

    def __eq__(self, other: "TensorStatisticObservationPoint"):
        return self.target_point == other.target_point


def create_register_input_hook(collector: TensorCollector) -> Callable[[torch.Tensor], torch.Tensor]:
    """
    Function to create register inputs hook function.

    :param collector: Collector to use in resulting hook.
    :return: Register inputs hook function.
    """

    def register_inputs_hook(x: Union[torch.Tensor, tuple]) -> torch.Tensor:
        """
        Register inputs hook function.

        :parameter x: tensor to register in hook.
        :return: tensor to register in hook.
        """
        with no_nncf_trace():
            x_unwrapped = maybe_get_values_from_torch_return_type(x)
            collector.register_input_for_all_reducers(Tensor(x_unwrapped))
        return x

    return register_inputs_hook


class TensorStatisticsCollectionBuilder(PTCompressionAlgorithmBuilder):
    def __init__(
        self,
        config: NNCFConfig,
        observation_points_vs_collectors: Dict[TensorStatisticObservationPoint, TensorStatisticCollectorBase],
    ):
        super().__init__(config)
        self._observation_points_vs_collectors = observation_points_vs_collectors

    def _get_transformation_layout(self, target_model: NNCFNetwork) -> PTTransformationLayout:
        # Will it really suffice to use a single collector for all threads? After all, each of the threads
        # receives its own data, and should we use a thread-local collector, there would have to be a
        # separate thread reduction step involved. Still, is there a better option here than to rely on GIL?
        layout = PTTransformationLayout()
        for op, rs_vs_collector in self._observation_points_vs_collectors.items():
            for collector in rs_vs_collector.values():
                command = PTInsertionCommand(
                    op.target_point,
                    create_register_input_hook(collector=collector),
                    TransformationPriority.FP32_TENSOR_STATISTICS_OBSERVATION,
                )
                layout.register(command)
        return layout

    def _build_controller(self, model: NNCFNetwork) -> "TensorStatisticsCollectionController":
        return TensorStatisticsCollectionController(
            model, {k.target_point: v for k, v in self._observation_points_vs_collectors.items()}
        )

    def _handle_frozen_layers(self, target_model: NNCFNetwork):
        pass

    def initialize(self, model: NNCFNetwork) -> None:
        pass

    def _get_algo_specific_config_section(self) -> Dict:
        return {}


class TensorStatisticsCollectionController(PTCompressionAlgorithmController):
    def __init__(
        self, target_model: NNCFNetwork, ip_vs_collector_dict: Dict[PTTargetPoint, TensorStatisticCollectorBase]
    ):
        super().__init__(target_model)
        self.ip_vs_collector_dict = ip_vs_collector_dict
        self._scheduler = StubCompressionScheduler()
        self._loss = ZeroCompressionLoss("cpu")

    @property
    def loss(self) -> ZeroCompressionLoss:
        return self._loss

    @property
    def scheduler(self) -> StubCompressionScheduler:
        return self._scheduler

    def start_collection(self):
        for collector in self.ip_vs_collector_dict.values():
            collector.enable()

    def stop_collection(self):
        for collector in self.ip_vs_collector_dict.values():
            collector.disable()

    def compression_stage(self) -> CompressionStage:
        return CompressionStage.FULLY_COMPRESSED

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        return NNCFStatistics()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
from functools import cmp_to_key
from functools import partial
from operator import itemgetter
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

import networkx as nx
import torch

from nncf.common.graph.definitions import MODEL_OUTPUT_OP_NAME
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNodeName
from nncf.common.logging import nncf_logger
from nncf.experimental.torch.search_building_blocks.search_graph import SearchGraph
from nncf.experimental.torch.search_building_blocks.search_graph import SearchGraphNode
from nncf.experimental.torch.search_building_blocks.search_graph import (
    check_graph_has_no_act_layer_duplication_after_block_removal,
)
from nncf.experimental.torch.search_building_blocks.search_graph import (
    check_graph_has_no_duplicate_edges_after_block_removal,
)
from nncf.experimental.torch.search_building_blocks.search_graph import (
    check_graph_has_no_hanging_edges_after_block_removal,
)
from nncf.experimental.torch.search_building_blocks.search_graph import get_num_ops_in_block
from nncf.experimental.torch.search_building_blocks.search_graph import get_search_graph
from nncf.torch.dynamic_graph.operation_address import OperationAddress
from nncf.torch.graph.graph import PTNNCFGraph
from nncf.torch.graph.operator_metatypes import PTDropoutMetatype
from nncf.torch.graph.operator_metatypes import PTLinearMetatype
from nncf.torch.graph.operator_metatypes import PTMatMulMetatype
from nncf.torch.layers import NNCF_MODULES_OP_NAMES
from nncf.torch.nncf_network import NNCFNetwork

IgnoredNameOperators = [*PTDropoutMetatype.get_all_aliases(), MODEL_OUTPUT_OP_NAME]
OrdinalIDs = List[List[int]]
GroupedBlockIDs = Dict[int, List[int]]

ShapeVsNodesMap = Dict[str, Set[SearchGraphNode]]


class PotentialBuildingBlock:
    """
    Describes a building block that is uniquely defined by the first skipped node and end nodes.
    """

    def __init__(self, first_skipped_node: SearchGraphNode, end_node: SearchGraphNode):
        self.first_skipped_node = first_skipped_node
        self.end_node = end_node

    def __eq__(self, __o: "PotentialBuildingBlock") -> bool:
        return self.first_skipped_node == __o.first_skipped_node and self.end_node == __o.end_node


class BuildingBlock:
    """
    Describes a building block that is uniquely defined by the start and end nodes.
    """

    def __init__(self, start_node_name: NNCFNodeName, end_node_name: NNCFNodeName):
        self.start_node_name = start_node_name
        self.end_node_name = end_node_name

    def __eq__(self, __o: "BuildingBlock") -> bool:
        return self.start_node_name == __o.start_node_name and self.end_node_name == __o.end_node_name

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return "[START NODE: {}, END_NODE: {}]".format(self.start_node_name, self.end_node_name)

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {"start_node_name": self.start_node_name, "end_node_name": self.end_node_name}

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "BuildingBlock":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return BuildingBlock(**state)


class BuildingBlockType(Enum):
    """
    Describes type of building block for transformers-based network.
    `MHSA` type is characterized by the presence 4 FC and 2 MatMul layers.
    `FF` type is characterized by the presence 2 FC layers.
    """

    MHSA = "MHSA"
    FF = "FF"
    Unknown = "unknown"

    def __eq__(self, other: "BuildingBlockType"):
        return self.__dict__ == other.__dict__

    def __hash__(self) -> int:
        return hash(self.name)


class EBBlocksStateNames:
    BASIC_BLOCK = "basic_block"
    BLOCK_TYPE = "block_type"
    ORDINAL_IDS = "ordinal_ids"
    OP_ADDRESSES = "op_addresses"


class ExtendedBuildingBlock:
    """
    Provides extended information about building block. In addition to the addresses of boundary operations, it defines
    block type, indexes of start and end node and addresses all operations inside the block.
    """

    _state_names = EBBlocksStateNames

    def __init__(
        self,
        basic_block: BuildingBlock,
        block_type: BuildingBlockType,
        ordinal_ids: Optional[List[int]],
        op_addresses: Optional[Set[OperationAddress]],
    ):
        self.basic_block = basic_block
        self.block_type = block_type
        self.ordinal_ids = ordinal_ids
        self.op_addresses = op_addresses

    def __str__(self) -> str:
        return "[START NODE: {}, END_NODE: {}] [{}, {}] #ops={}".format(
            self.basic_block.start_node_name, self.basic_block.end_node_name, *self.ordinal_ids, len(self.op_addresses)
        )

    @property
    def start_node_name(self):
        return self.basic_block.start_node_name

    @property
    def end_node_name(self):
        return self.basic_block.end_node_name

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "ExtendedBuildingBlock":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        bbtype = BuildingBlockType(state[cls._state_names.BLOCK_TYPE])
        bblock = BuildingBlock.from_state(state[cls._state_names.BASIC_BLOCK])
        op_addresses = {
            OperationAddress.from_str(op_address_state) for op_address_state in state[cls._state_names.OP_ADDRESSES]
        }
        ordinal_ids = state[cls._state_names.ORDINAL_IDS]
        kwargs = {
            cls._state_names.BLOCK_TYPE: bbtype,
            cls._state_names.BASIC_BLOCK: bblock,
            cls._state_names.OP_ADDRESSES: op_addresses,
            cls._state_names.ORDINAL_IDS: ordinal_ids,
        }
        return cls(**kwargs)

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        op_addresses = sorted([str(op_address) for op_address in self.op_addresses])
        return {
            self._state_names.BLOCK_TYPE: self.block_type.name,
            self._state_names.BASIC_BLOCK: self.basic_block.get_state(),
            self._state_names.OP_ADDRESSES: op_addresses,
            self._state_names.ORDINAL_IDS: self.ordinal_ids,
        }

    def __eq__(self, other: "ExtendedBuildingBlock"):
        return (
            self.block_type == other.block_type
            and self.basic_block == other.basic_block
            and self.op_addresses == other.op_addresses
            and self.ordinal_ids == other.ordinal_ids
        )


BuildingBlocks = List[BuildingBlock]
ExtendedBuildingBlocks = List[ExtendedBuildingBlock]


def add_node_to_aux_struct(node: SearchGraphNode, shape: List[int], shape_map: ShapeVsNodesMap):
    """
    Add to shape_map key of node for corresponds shape.
    """
    str_shape = str(shape)
    if str_shape in shape_map:
        shape_map[str_shape].add(node)
    else:
        shape_map[str_shape] = {node}


def remove_duplicates(sorted_blocks: List[PotentialBuildingBlock]) -> List[PotentialBuildingBlock]:
    """
    Removes identical blocks - that have the same nodes on boundaries
    :param sorted_blocks: a sorted list of block. Some determined order is required as soon as otherwise filtration
    may lead to a different result.
    :return: filtered list of blocks without duplicates
    """
    id_pairs = set()
    filtered_blocks = []
    for b in sorted_blocks:
        current_pair_ids = (b.first_skipped_node.main_id, b.end_node.main_id)
        if current_pair_ids not in id_pairs:
            filtered_blocks.append(b)
        id_pairs.add(current_pair_ids)
    return filtered_blocks


def compare_for_building_block(a: PotentialBuildingBlock, b: PotentialBuildingBlock):
    """
    Orders the blocks in ascending order of the end node index.
    If the indices of the end nodes are the same, the blocks are ordered by the
    index of the first skipped node.Otherwise - by number of operations in the block.
    """
    if a.end_node.bottom_id != b.end_node.bottom_id:
        return a.end_node.bottom_id - b.end_node.bottom_id
    if a.first_skipped_node.main_id != b.first_skipped_node.main_id:
        return a.first_skipped_node.main_id - b.first_skipped_node.main_id
    num_ops_a = get_num_ops_in_block(a.first_skipped_node, a.end_node)
    num_ops_b = get_num_ops_in_block(b.first_skipped_node, b.end_node)
    return num_ops_a - num_ops_b


def get_building_block_for_original_graph(
    building_blocks: List[PotentialBuildingBlock], orig_graph: PTNNCFGraph
) -> ExtendedBuildingBlocks:
    """
    Restore the original names and ids of the start and end of the block in original graph.
    Very cost expensive function, because of finding all op addresses in the block via nx.all_simple_paths function.
    """
    building_block_in_orig_format = []
    for block in building_blocks:
        id_end = block.end_node.bottom_id
        start_node_id = get_start_node_id(block, orig_graph)
        block_in_orig_format = BuildingBlock(
            orig_graph.get_node_key_by_id(start_node_id).split(" ")[-1],
            orig_graph.get_node_key_by_id(id_end).split(" ")[-1],
        )
        ordinal_ids = [start_node_id, id_end]
        op_addresses = get_all_node_op_addresses_in_block(orig_graph, block_in_orig_format)
        block_type = get_type_building_block(op_addresses)
        ext_block = ExtendedBuildingBlock(block_in_orig_format, block_type, ordinal_ids, op_addresses)
        building_block_in_orig_format.append(ext_block)
    return building_block_in_orig_format


def get_potential_candidate_for_block(search_graph: SearchGraph) -> Tuple[ShapeVsNodesMap, ShapeVsNodesMap]:
    """
    Distributes all nodes to the same output and input shapes.

    :param search_graph: A wrapper over the graph of target model, which represents the DNN execution graph transformed
    by pattern matching, merging nodes and inserting auxiliary nodes
    :return: two dictionaries that represents all input/output shapes (in a string form) of operations in the model and
    maps these shapes to nodes that have such shapes on input or output correspondingly.
    """
    act_input_shape = {}  # key - str(shape), value - set of node_keys
    act_output_shape = {}  # key - str(shape), value - set of node_keys
    for node in search_graph.get_all_nodes():
        next_edges = search_graph.get_next_edges(node.node_key)
        prev_edges = search_graph.get_prev_edges(node.node_key)
        for _, edge_attr in next_edges.items():
            search_graph.set_node_attr(
                node.node_key, SearchGraph.ACTIVATION_OUTPUT_SHAPE_ATTR, edge_attr[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR]
            )
            if not node.is_dummy:
                add_node_to_aux_struct(node, edge_attr[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR], act_output_shape)
            break
        for _, edge_attr in prev_edges.items():
            search_graph.set_node_attr(
                node.node_key, SearchGraph.ACTIVATION_OUTPUT_SHAPE_ATTR, edge_attr[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR]
            )
            add_node_to_aux_struct(node, edge_attr[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR], act_input_shape)
            break
    return act_input_shape, act_output_shape


def itemgetter_force_tuple(*indexes):
    """
    itemgetter wrapper that always returns a tuple. The original function may return both: iterable and a single
    non-iterable element, which is not convenient in the general case.
    """
    getter = itemgetter(*indexes)
    if len(indexes) == 1:
        return lambda seq: (getter(seq),)  # Wrap in a tuple.
    return getter


def is_target_block_type(
    start_node_id: int,
    end_node_id: int,
    orig_graph: NNCFGraph,
    target_block_types: Optional[List[BuildingBlockType]] = None,
) -> bool:
    """
    Returns true if block has a type equal to one of the specified.
    :param start_node_id: bottom index of the starting node in search graph
    :param end_node_id: bottom index of the ending node in search graph
    :param orig_graph: original non-compressed graph.
    :param target_block_types: list of block types to match the type of the given block
    :return: additional info for the building blocks - block type and addresses of all operations inside the block .
    """
    if target_block_types is None:
        return True
    block_for_ops = BuildingBlock(
        orig_graph.get_node_key_by_id(start_node_id).split(" ")[-1],
        orig_graph.get_node_key_by_id(end_node_id).split(" ")[-1],
    )
    op_addresses = get_all_node_op_addresses_in_block(orig_graph, block_for_ops)
    block_type = get_type_building_block(op_addresses)
    return block_type in target_block_types


class BlockFilteringStrategy(Enum):
    """
    Defines strategy for filtering overlapping blocks.
    KEEP_SMALL - gives a preference to small blocks. It starts from the smallest block and filters all blocks that
    intersect or include it. Then it finds the next smallest block from the remaining ones and repeats the procedure.
    KEEP_SEQUENTIAL - gives a preference to sequential blocks, which follow each other in the model. This strategy
    is helpful for Progressive Shrinking Algorithm.
    """

    KEEP_SMALL = "keep_small"
    KEEP_SEQUENTIAL = "keep_sequential"


def get_building_blocks(
    compressed_model: NNCFNetwork,
    max_block_size: int = 50,
    min_block_size: int = 5,
    block_filter_strategy=BlockFilteringStrategy.KEEP_SEQUENTIAL,
    hw_fused_ops: bool = True,
    target_block_types: Optional[List[BuildingBlockType]] = None,
) -> Tuple[ExtendedBuildingBlocks, GroupedBlockIDs]:
    """
    This algorithm finds building blocks based on the analysis of the transformed graph.
    A building block is a block that satisfies the following rules:
    - has one input and one output tensors
    - input and output tensors shape are the same
    - removing a block from the graph (that is, the layers included in the block are not executed)
      does not lead to duplication of edges along which the same tensor flows
    - removing a block from the graph (that is, the layers included in the block are not executed)
      does not lead to dangling edges
    - removing a block from the graph (that is, the layers included in the block are not executed)
      does not lead to duplicate activation layers
    """
    if min_block_size > max_block_size:
        raise AttributeError(
            f"Minimal value for block size {min_block_size} can not be more than maximum one "
            f"{max_block_size}. Change max_block_size or min_block_size."
        )
    orig_graph = compressed_model.nncf.get_original_graph()  # PTNNCFGraph
    blocks = get_potential_building_blocks(orig_graph, hw_fused_ops, min_block_size, max_block_size)
    sorted_blocks = sorted(blocks, key=cmp_to_key(compare_for_building_block))
    filtered_building_blocks = remove_duplicates(sorted_blocks)
    filtered_building_blocks = remove_overlapping_blocks(
        filtered_building_blocks, block_filter_strategy, orig_graph, target_block_types
    )
    ext_building_blocks = get_building_block_for_original_graph(filtered_building_blocks, orig_graph)
    filtered_basic_blocks = [eb.basic_block for eb in ext_building_blocks]
    group_dependent = get_group_of_dependent_blocks(filtered_basic_blocks)
    return ext_building_blocks, group_dependent


def get_potential_building_blocks(orig_graph: NNCFGraph, hw_fused_ops: bool, min_block_size: int, max_block_size: int):
    """
    Builds a list of potential building blocks that has the same shapes on input and output and that satisfy some
    heuristic rules.
    :param orig_graph: NNCF graph of the non-compressed model
    :param hw_fused_ops: If True, automatic block search will not relate operations, which are fused on inference,
         into different blocks for skipping.
    :param min_block_size: minimum number of operations that should be in the block. Blocks with smaller number of
    operations are not considered.
    :param max_block_size: maximum number of operations that should be in the block. Blocks with larger number of
    operations are not considered.
    :return: list of potential building blocks
    """
    sgraph = get_search_graph(orig_graph, hw_fused_ops)
    heuristic_rules = [
        check_graph_has_no_duplicate_edges_after_block_removal,
        check_graph_has_no_act_layer_duplication_after_block_removal,
        check_graph_has_no_hanging_edges_after_block_removal,
    ]
    blocks = []
    act_input_shape, act_output_shape = get_potential_candidate_for_block(sgraph)
    for shape, first_skipped_nodes in act_input_shape.items():
        for first_skipped_node in first_skipped_nodes:
            previous_nodes = sgraph.get_prev_nodes(first_skipped_node.node_key)
            if first_skipped_node.node_type in IgnoredNameOperators or len(previous_nodes) != 1:
                continue
            for end_node in act_output_shape[shape]:
                if end_node.main_id <= first_skipped_node.main_id:
                    continue
                if end_node.node_type in IgnoredNameOperators:
                    continue
                num_ops_in_block = get_num_ops_in_block(first_skipped_node, end_node)
                if num_ops_in_block > max_block_size or num_ops_in_block < min_block_size:
                    continue
                # CHECK RULES
                all_rules_is_true = True
                for rule_fn in heuristic_rules:
                    if not rule_fn(sgraph, first_skipped_node, end_node):
                        all_rules_is_true = False
                        break
                if all_rules_is_true:
                    blocks.append(PotentialBuildingBlock(first_skipped_node, end_node))
    if len(blocks) > 300:
        nncf_logger.warning(
            "Number of potential building blocks is too much. The processing time can be high. "
            "Shallow the accepted range for the length of building blocks via "
            "max_block_size and min_block_size to accelerate the search process."
        )
    return blocks


def remove_overlapping_blocks(
    building_blocks: List[PotentialBuildingBlock],
    block_filter_strategy: BlockFilteringStrategy,
    orig_graph: NNCFGraph,
    target_block_types: List[BuildingBlockType],
) -> List[PotentialBuildingBlock]:
    """
    Applies the given filtering strategy to remove overlapping blocks - pairs of blocks that shares common operations.
    :param building_blocks: list of potential building blocks
    :param block_filter_strategy: strategy to remove overlapping blocks
    :param orig_graph: NNCF graph for non-compressed model
    :param target_block_types: list of target block types that shouldn't be removed
    :return: list of non-overlapping building blocks
    """
    start_ids = []
    end_ids = []
    num_ops_in_blocks = []
    for block in building_blocks:
        start_node_id = get_start_node_id(block, orig_graph)
        id_end = block.end_node.bottom_id
        num_ops_in_block = id_end - start_node_id - 1
        start_ids.append(start_node_id)
        end_ids.append(id_end)
        num_ops_in_blocks.append(num_ops_in_block)
    is_target_block_type_fn = partial(
        is_target_block_type, orig_graph=orig_graph, target_block_types=target_block_types
    )
    get_indexes_of_overlapping_blocks_fn = get_indexes_of_overlapping_blocks_seq
    if block_filter_strategy == BlockFilteringStrategy.KEEP_SMALL:
        get_indexes_of_overlapping_blocks_fn = get_indexes_of_overlapping_blocks_min
    ids_of_overlapping_blocks = get_indexes_of_overlapping_blocks_fn(
        start_ids=start_ids,
        end_ids=end_ids,
        num_ops_in_blocks=num_ops_in_blocks,
        is_target_block_type_fn=is_target_block_type_fn,
    )
    if ids_of_overlapping_blocks:
        building_blocks = [block for i, block in enumerate(building_blocks) if i not in ids_of_overlapping_blocks]
    return building_blocks


def get_start_node_id(block: PotentialBuildingBlock, orig_graph: NNCFGraph) -> int:
    """
    Returns id of the starting node of the block - the node right before first skipped node.
    :param block: potential building block
    :param orig_graph: original nncf graph
    :return: id of starting node
    """
    id_st = block.first_skipped_node.main_id
    first_skipped_node = orig_graph.get_node_by_id(id_st)
    input_nodes = orig_graph.get_input_nodes()
    start_node_id = id_st
    if first_skipped_node not in input_nodes and not block.first_skipped_node.is_dummy:
        previous_nodes = orig_graph.get_previous_nodes(first_skipped_node)
        num_inputs = len(previous_nodes)
        assert num_inputs == 1, f"building block should have a single input, but it has {num_inputs} inputs."
        start_node_id = previous_nodes[0].node_id
    return start_node_id


def get_indexes_of_overlapping_blocks_seq(
    start_ids: List[int],
    end_ids: List[int],
    num_ops_in_blocks: List[int],
    is_target_block_type_fn: Optional[Callable[[int, int], bool]] = None,
) -> Set[int]:
    """
    The function takes coordinates of the building block (start and end ids) and finds indexes of overlapping blocks.
    After only disjoint blocks remain after filtering the found blocks.
    :param start_ids: indexes of start node in the blocks.
    :param end_ids: indexes of end node in the blocks.
    :param num_ops_in_blocks: number of operations in the block.
    :param is_target_block_type_fn: functor that defines whether the block of target type by taking indexes of start
    and end node. Currently, types are ignored in this strategy
    :return: set of indexes of the overlapping blocks
    """
    num_blocks = len(num_ops_in_blocks)
    block_graph = nx.DiGraph()
    for i, (s, e, n) in enumerate(zip(start_ids, end_ids, num_ops_in_blocks)):
        block_graph.add_edge(s, e, attr={"cost": 4 - n, "block_id": i})

    result = set(range(num_blocks))
    ids_of_not_overlapping_blocks = set()
    while block_graph.nodes:
        ids_of_not_overlapping_nodes = nx.dag_longest_path(block_graph, weight="cost")
        node_ids_to_remove = set()
        for i in range(len(ids_of_not_overlapping_nodes) - 1):
            data = block_graph.get_edge_data(ids_of_not_overlapping_nodes[i], ids_of_not_overlapping_nodes[i + 1])
            ids_of_not_overlapping_blocks.add(data["attr"]["block_id"])
        for node_id in ids_of_not_overlapping_nodes:
            block_graph.remove_node(node_id)

        left_border = ids_of_not_overlapping_nodes[0]
        right_border = ids_of_not_overlapping_nodes[-1]
        for node_id1, node_id2, data in block_graph.edges(data=True):
            i = data["attr"]["block_id"]
            does_intersect_found_block = (
                left_border < start_ids[i] < right_border or left_border < end_ids[i] < right_border
            )
            does_include_found_block = (
                start_ids[i] <= left_border <= end_ids[i] and start_ids[i] <= right_border <= end_ids[i]
            )
            if does_intersect_found_block or does_include_found_block:
                node_ids_to_remove.add(node_id1)
                node_ids_to_remove.add(node_id2)

        for node_id in node_ids_to_remove:
            block_graph.remove_node(node_id)

    return result - ids_of_not_overlapping_blocks


def get_indexes_of_overlapping_blocks_min(
    start_ids: List[int],
    end_ids: List[int],
    num_ops_in_blocks: List[int],
    is_target_block_type_fn: Optional[Callable[[int, int], bool]] = None,
) -> Set[int]:
    """
    The function takes coordinates of the building block (start and end ids) and finds indexes of overlapping blocks.
    After only disjoint blocks remain after filtering the found blocks.
    :param start_ids: indexes of start node in the blocks.
    :param end_ids: indexes of end node in the blocks.
    :param num_ops_in_blocks: number of operations in the block.
    :param is_target_block_type_fn: functor that defines whether the block of target type by taking indexes of start
    and end node.
    :return: set of indexes of the overlapping blocks
    """
    if is_target_block_type_fn is None:
        is_target_block_type_fn = lambda *_: True
    result = set()
    if not start_ids or not end_ids or not num_ops_in_blocks:
        return result
    num_blocks = len(num_ops_in_blocks)

    indexes_to_sort, _ = zip(*sorted(enumerate(num_ops_in_blocks), key=itemgetter(1)))
    list_of_all_block_indexes = list(range(num_blocks))
    sorted_block_indexes = itemgetter_force_tuple(*indexes_to_sort)(list_of_all_block_indexes)

    curr_index = -1
    while list_of_all_block_indexes:
        is_found = False
        while not is_found and curr_index < num_blocks - 1:
            curr_index += 1
            found_block_id = sorted_block_indexes[curr_index]
            found_next_not_removed_block = found_block_id in list_of_all_block_indexes
            is_target_type = is_target_block_type_fn(start_ids[found_block_id], end_ids[found_block_id])
            if found_next_not_removed_block and is_target_type:
                is_found = True
            else:
                result.add(found_block_id)
            if found_block_id in list_of_all_block_indexes:
                list_of_all_block_indexes.remove(found_block_id)
        left_border = start_ids[found_block_id]
        right_border = end_ids[found_block_id]
        ids_to_remove = []
        for i in list_of_all_block_indexes:
            does_intersect_found_block = (
                left_border < start_ids[i] < right_border or left_border < end_ids[i] < right_border
            )
            does_include_found_block = (
                start_ids[i] <= left_border <= end_ids[i] and start_ids[i] <= right_border <= end_ids[i]
            )
            if does_intersect_found_block or does_include_found_block:
                ids_to_remove.append(i)
                result.add(i)
        for i in ids_to_remove:
            list_of_all_block_indexes.remove(i)
    return result


def get_group_of_dependent_blocks(blocks: BuildingBlocks) -> GroupedBlockIDs:
    """
    Building blocks can be categorized into groups. Blocks that follow each other in the graph
    (that is, they are connected by one edge) belong to the same group.

    :param: List of building blocks.
    :return: Dictionary where key is block index, value is group index.
    """
    if not blocks:
        return {}
    idx = 0
    groups = {idx: []}
    for i in range(len(blocks) - 1):
        next_start_node_name = blocks[i + 1].start_node_name
        curr_end_node_name = blocks[i].end_node_name
        if next_start_node_name == curr_end_node_name:
            groups[idx].append(i)
        else:
            groups[idx].append(i)
            idx += 1
            groups[idx] = []
    groups[idx].append(len(blocks) - 1)

    return groups


def get_all_node_op_addresses_in_block(graph: NNCFGraph, block: BuildingBlock) -> Set[OperationAddress]:
    """
    Returns set of operation addresses of all layers included in the block.

    :param graph: original non-compressed graph.
    :param block: Building blocks.
    :return: Set of operation addresses for building block.
    """
    simple_paths = graph.get_all_simple_paths(block.start_node_name, block.end_node_name)
    op_addresses = set()
    for node_keys_in_path in simple_paths:
        for node_key in node_keys_in_path:
            node = graph.get_node_by_key(node_key)
            op_addresses.add(OperationAddress.from_str(node.node_name))
    start_op_address = OperationAddress.from_str(block.start_node_name)
    op_addresses.remove(start_op_address)
    return op_addresses


def get_all_modules_in_blocks(
    compressed_model: NNCFNetwork, op_adresses_in_blocks: Set[OperationAddress]
) -> List[torch.nn.Module]:
    """
    Returns set of all modules included in the block.

    :param compressed_model: Target model.
    :param op_adresses_in_blocks: Set of operation addresses for building block.
    :return: List of module for building block.
    """
    modules = []
    for op_address in op_adresses_in_blocks:
        if op_address.operator_name in NNCF_MODULES_OP_NAMES:
            modules.append(compressed_model.nncf.get_module_by_scope(op_address.scope_in_model))
    return modules


def get_type_building_block(op_addresses_in_block: Set[OperationAddress]) -> BuildingBlockType:
    """
    Returns type of building block.
    """
    count_matmul = 0
    count_fc = 0
    for op_address in op_addresses_in_block:
        if op_address.operator_name in PTMatMulMetatype.get_all_aliases():
            count_matmul += 1
        if op_address.operator_name in PTLinearMetatype.get_all_aliases():
            count_fc += 1
    if count_fc == 4 and count_matmul == 2:
        return BuildingBlockType.MHSA
    if count_fc == 2 and count_matmul == 0:
        return BuildingBlockType.FF
    return BuildingBlockType.Unknown



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Tuple, Union

import torch
from torch import nn

from nncf.common.engine import Engine
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend


class PTEngine(Engine):
    """
    Engine for the Pytorch backend.
    """

    def __init__(self, model: nn.Module):
        """
        Constructor.

        :param model: Pytorch module to infer.
        """

        self._model = model
        if get_backend(model) == BackendType.TORCH:
            self._model.eval()

    def infer(
        self, input_data: Union[torch.Tensor, Tuple[torch.Tensor], Dict[str, torch.Tensor]]
    ) -> Union[torch.Tensor, Dict[str, Any]]:
        """
        Runs Torch model on the provided input.

        :param input_data: Inputs for the model.
        :return: Model outputs.
        """

        if isinstance(input_data, dict):
            return self._model(**input_data)
        if isinstance(input_data, tuple):
            return self._model(*input_data)
        return self._model(input_data)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path as osp
from typing import Dict

import torch
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torchvision.transforms import ToTensor

from nncf.common.accuracy_aware_training.runner import BaseAccuracyAwareTrainingRunner
from nncf.common.accuracy_aware_training.runner import BaseAdaptiveCompressionLevelTrainingRunner
from nncf.torch.accuracy_aware_training.utils import is_main_process
from nncf.torch.checkpoint_loading import load_state

try:
    from torch.utils.tensorboard import SummaryWriter

    TENSORBOARD_AVAILABLE = True
except ImportError:
    TENSORBOARD_AVAILABLE = False


class PTAccuracyAwareTrainingRunner(BaseAccuracyAwareTrainingRunner):
    """
    BaseAccuracyAwareTrainingRunner
    The Training Runner implementation for PyTorch training code.
    """

    def initialize_logging(self, log_dir=None, tensorboard_writer=None):
        if not is_main_process():
            return
        # Only the main process should initialize and create a log directory, other processes don't use it
        super().initialize_logging(log_dir, tensorboard_writer)
        if self._tensorboard_writer is None and TENSORBOARD_AVAILABLE:
            self._tensorboard_writer = SummaryWriter(self._log_dir)

    def validate(self, model):
        with torch.no_grad():
            self.current_val_metric_value = self._validate_fn(model, epoch=self.cumulative_epoch_count)
        is_best = (not self.is_higher_metric_better) != (self.current_val_metric_value > self.best_val_metric_value)
        if is_best:
            self.best_val_metric_value = self.current_val_metric_value
        return self.current_val_metric_value

    def dump_statistics(self, model, compression_controller):
        if not is_main_process():
            return
        super().dump_statistics(model, compression_controller)

    def update_learning_rate(self):
        if self._update_learning_rate_fn is not None:
            self._update_learning_rate_fn(
                self.lr_scheduler, self.training_epoch_count, self.current_val_metric_value, self.current_loss
            )
        else:
            if self.lr_scheduler is not None and self.lr_updates_needed:
                self.lr_scheduler.step(
                    self.training_epoch_count
                    if not isinstance(self.lr_scheduler, ReduceLROnPlateau)
                    else self.best_val_metric_value
                )

    def reset_training(self):
        self.configure_optimizers()

        optimizers = self.optimizer if isinstance(self.optimizer, (tuple, list)) else [self.optimizer]
        for optimizer in optimizers:
            for param_group in optimizer.param_groups:
                param_group["lr"] *= self.base_lr_reduction_factor_during_search

        lr_schedulers = self.lr_scheduler if isinstance(self.lr_scheduler, (tuple, list)) else [self.lr_scheduler]
        for lr_scheduler in lr_schedulers:
            if lr_scheduler is None:
                continue
            for attr_name in ["base_lrs", "init_lr"]:
                if hasattr(lr_scheduler, attr_name):
                    setattr(
                        lr_scheduler,
                        attr_name,
                        [
                            base_lr * self.base_lr_reduction_factor_during_search
                            for base_lr in getattr(lr_scheduler, attr_name)
                        ],
                    )

        self.training_epoch_count = 0
        self.best_val_metric_value = 0
        self.current_val_metric_value = 0

    def dump_checkpoint(self, model, compression_controller):
        if not is_main_process():
            return
        super().dump_checkpoint(model, compression_controller)

    def load_best_checkpoint(self, model):
        if not is_main_process():
            return
        super().load_best_checkpoint(model)

    def _save_checkpoint(self, model, compression_controller, checkpoint_path):
        optimizers = self.optimizer if isinstance(self.optimizer, (tuple, list)) else [self.optimizer]
        checkpoint = {
            "epoch": self.cumulative_epoch_count + 1,
            "state_dict": model.state_dict(),
            "compression_state": compression_controller.get_compression_state(),
            "best_metric_val": self.best_val_metric_value,
            "current_val_metric_value": self.current_val_metric_value,
            "optimizer": [optimizer.state_dict() for optimizer in optimizers],
        }
        torch.save(checkpoint, checkpoint_path)

    def _load_checkpoint(self, model, checkpoint_path):
        if self._load_checkpoint_fn is not None:
            self._load_checkpoint_fn(model, checkpoint_path)
        else:
            resuming_checkpoint = torch.load(checkpoint_path, map_location="cpu")
            resuming_model_state_dict = resuming_checkpoint.get("state_dict", resuming_checkpoint)
            load_state(model, resuming_model_state_dict, is_resume=True)

    def _make_checkpoint_path(self, is_best, compression_rate=None):
        extension = ".pth"
        return osp.join(self._checkpoint_save_dir, f'acc_aware_checkpoint_{"best" if is_best else "last"}{extension}')

    def add_tensorboard_scalar(self, key, data, step):
        if is_main_process() and self.verbose and self._tensorboard_writer is not None:
            self._tensorboard_writer.add_scalar(key, data, step)

    def add_tensorboard_image(self, key, data, step):
        if is_main_process() and self.verbose and self._tensorboard_writer is not None:
            self._tensorboard_writer.add_image(key, ToTensor()(data), step)


class PTAdaptiveCompressionLevelTrainingRunner(
    BaseAdaptiveCompressionLevelTrainingRunner, PTAccuracyAwareTrainingRunner
):
    def __init__(
        self,
        accuracy_aware_training_params: Dict,
        uncompressed_model_accuracy: float,
        verbose: bool = True,
        dump_checkpoints: bool = True,
        lr_updates_needed: bool = True,
        minimal_compression_rate: float = 0.0,
        maximal_compression_rate: float = 0.95,
    ):
        super().__init__(
            accuracy_aware_training_params,
            uncompressed_model_accuracy,
            verbose,
            dump_checkpoints,
            lr_updates_needed,
            minimal_compression_rate=minimal_compression_rate,
            maximal_compression_rate=maximal_compression_rate,
        )

    def _make_checkpoint_path(self, is_best, compression_rate=None):
        extension = ".pth"
        base_path = osp.join(self._checkpoint_save_dir, "acc_aware_checkpoint")
        if is_best:
            if compression_rate is None:
                raise ValueError("Compression rate cannot be None")
            return f"{base_path}_best_{compression_rate:.3f}{extension}"
        return f"{base_path}_last{extension}"
